{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Author : Luc DAVID Email: luckylazarus@free.fr
    Copyright (C) 2007
    Last update : 21/12/2007

  --------------------------------------------------------------------------- }

unit SqlitePassCustomFieldDefsDialog;
{$IFDEF FPC}
 {$mode DELPHI}
{$ENDIF}

interface

uses
 {$IFDEF FPC}
  LResources,
 {$ENDIF}
 {$IFDEF WIN32}
  Windows,
 {$ENDIF}
  Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Math, Buttons, ExtCtrls, DB, ComCtrls, SqlitePassDbo, SqlitePassUtils;

type
  TSqlitePassCustomFieldDefsDlg = class(TForm)
    PanelBottom: TPanel;
    PanelDatatypes: TPanel;
    BtOk: TButton;
    FieldDefsScrollBox: TScrollBox;
    HeaderControl1: THeaderControl;
    SbDelete: TSpeedButton;
    SbAdd: TSpeedButton;
    BtnSaveToDatabase: TButton;
    BtnLoadFromDatabase: TButton;
    BtnCancel: TButton;
    BtClearAll: TButton;
    procedure SbAddClick(Sender: TObject);
    procedure SbDeleteClick(Sender: TObject);
    procedure BtOkClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure BtnLoadFromDatabaseClick(Sender: TObject);
    procedure BtnSaveToDatabaseClick(Sender: TObject);
    procedure BtClearAllClick(Sender: TObject);
  private
    Db: TSqlitePassDatabase;
    RowTop: Integer;
    CbTableNameComponents: TList;
    CbFieldNameComponents: TList;
    CbFieldTypeComponents: TList;
    EdFieldSizeComponents: Tlist;
    EdFieldPrecisionComponents: Tlist;
    CurrentRow: Integer;
    FDatabaseTypeOptions: TSqlitePassDatabaseDataTypeOptions;
    Procedure AddNewCustomFieldDef;
    Procedure DeleteCustomFieldDef(Index: Integer);
    procedure DisplayCustomFieldDefs;
    Procedure OnCbTableNameChange(Sender: TObject);
    procedure UpdateFieldNames(CurrentCbFieldName: TComboBox; TableDef: TSqlitePassTableDef);
  public
    constructor Create(AOwner: TComponent; DatabaseTypeOptions: TSqlitePassDatabaseDataTypeOptions); reintroduce;
    destructor Destroy; override;
  end;

var
  SqlitePassCustomFieldDefsDlg: TSqlitePassCustomFieldDefsDlg;

implementation

{$IFNDEF FPC}
 {$R *.DFM}
{$ENDIF}

{ Constructor and Destructor }

constructor TSqlitePassCustomFieldDefsDlg.Create(AOwner: TComponent; DatabaseTypeOptions: TSqlitePassDatabaseDataTypeOptions);
begin
inherited Create(AOwner);
FDatabaseTypeOptions := DatabaseTypeOptions;
Db := TSqlitePassDatabase(FDatabaseTypeOptions.GetOwner);
RowTop := -16;
CbTableNameComponents := TList.Create;
CbFieldNameComponents := TList.Create;
CbFieldTypeComponents := TList.Create;
EdFieldSizeComponents := Tlist.Create;
EdFieldPrecisionComponents := Tlist.Create;
DisplayCustomFieldDefs;
end;

destructor TSqlitePassCustomFieldDefsDlg.Destroy;
begin
CbTableNameComponents.Free;
CbFieldNameComponents.Free;
CbFieldTypeComponents.Free;
EdFieldSizeComponents.Free;
EdFieldPrecisionComponents.Free;
inherited;
end;

procedure TSqlitePassCustomFieldDefsDlg.DisplayCustomFieldDefs;
var
i: Integer;
begin
Try
  FieldDefsScrollBox.Visible := False;
  For i := 0 to Pred(FDatabaseTypeOptions.CustomFieldDefs.Count) do
      begin
      AddNewCustomFieldDef;
      TComboBox(CbTableNameComponents[i]).ItemIndex := TComboBox(CbTableNameComponents[i]).Items.IndexOf(FDatabaseTypeOptions.CustomFieldDefs[i].TableName);
      UpdateFieldNames(TComboBox(CbFieldNameComponents[i]), Db.TableDefs.FindTable(TComboBox(CbTableNameComponents[i]).Text));
      TComboBox(CbFieldNameComponents[i]).ItemIndex := TComboBox(CbFieldNameComponents[i]).Items.IndexOf(FDatabaseTypeOptions.CustomFieldDefs[i].FieldName);
      TComboBox(CbFieldTypeComponents[i]).ItemIndex := TComboBox(CbFieldTypeComponents[i]).Items.IndexOf(FieldTypeToString(FDatabaseTypeOptions.CustomFieldDefs[i].FieldType));
      TEdit(EdFieldSizeComponents[i]).Text := IntToStr(FDatabaseTypeOptions.CustomFieldDefs[i].FieldSize);
      TEdit(EdFieldPrecisionComponents[i]).Text := IntToStr(FDatabaseTypeOptions.CustomFieldDefs[i].FieldPrecision);
      end;
  if CbTableNameComponents.Count = 0
     then AddNewCustomFieldDef;
Finally
  FieldDefsScrollBox.Visible := True;
end;
end;

procedure TSqlitePassCustomFieldDefsDlg.AddNewCustomFieldDef;
var
i: integer;
NewCbTableName: TComboBox;
NewCbFieldName: TComboBox;
NewCbFieldType: TComboBox;
NewEdFieldSize: TEdit;
NewEdFieldPrecision: TEdit;
ft: TFieldType;
begin
Inc(RowTop, 20);

NewCbTableName := TComboBox.Create(FieldDefsScrollBox);
With NewCbTableName do
     begin
     Parent := FieldDefsScrollBox;
     Text := '';
     Top  := RowTop;
     Left := 3;
     Width := 160;
     Height := 20;
     Style := csDropDownList;
     for i:=0 to Pred(Db.TableDefs.Count)
         do Items.Add(Db.TableDefs[i].TableName);

     Sorted := True;
     OnChange := OnCbTableNameChange;
     NewCbTableName.ItemIndex := -1;
     end;

NewCbFieldName := TComboBox.Create(FieldDefsScrollBox);
With NewCbFieldName do
     begin
     Parent := FieldDefsScrollBox;
     Top  := RowTop;
     Left := 171;
     Width := 155;
     Height := 20;
     Style := csDropDownList;
     Sorted := True;
     end;

NewCbFieldType := TComboBox.Create(FieldDefsScrollBox);
With NewCbFieldType do
     begin
     Parent := FieldDefsScrollBox;
     Top  := RowTop;
     Left := 332;
     Width := 155;
     Height := 20;
     For ft := Low(TFieldType) to High(TFieldType)
         do  Items.Add(FieldTypeToString(ft));

     Style := csDropDownList;
     Sorted := True;
     ItemIndex := 0;
     end;

NewEdFieldSize := TEdit.Create(FieldDefsScrollBox);
With NewEdFieldSize do
     begin
     Parent := FieldDefsScrollBox;
     Text := '';
     Top  := RowTop;
     Left := 495;
     Width := 55;
     Height := 20;
     Text := '0';
     end;

NewEdFieldPrecision := TEdit.Create(FieldDefsScrollBox);
With NewEdFieldPrecision do
     begin
     Parent := FieldDefsScrollBox;
     Text := '';
     Top  := RowTop;
     Left := 560;
     Width := 40;
     Height := 20;
     Text := '0';
     end;

CbTableNameComponents.Add(NewCbTableName);
CbFieldNameComponents.Add(NewCbFieldName);
CbFieldTypeComponents.Add(NewCbFieldType);
EdFieldSizeComponents.Add(NewEdFieldSize);
EdFieldPrecisionComponents.Add(NewEdFieldPrecision);
end;


procedure TSqlitePassCustomFieldDefsDlg.DeleteCustomFieldDef(Index: Integer);
begin
TComboBox(CbTableNameComponents[Index]).Free;
TComboBox(CbFieldNameComponents[Index]).Free;
TComboBox(CbFieldTypeComponents[Index]).Free;
TEdit(EdFieldSizeComponents[Index]).Free;
TEdit(EdFieldPrecisionComponents[Index]).Free;
CbTableNameComponents.Delete(Index);
CbFieldNameComponents.Delete(Index);
CbFieldTypeComponents.Delete(Index);
EdFieldSizeComponents.Delete(Index);
EdFieldPrecisionComponents.Delete(Index);
Dec(RowTop,20);
end;

{ Buttons Events }

procedure TSqlitePassCustomFieldDefsDlg.SbAddClick(Sender: TObject);
begin
AddNewCustomFieldDef;
end;

procedure TSqlitePassCustomFieldDefsDlg.SbDeleteClick(Sender: TObject);
begin
DeleteCustomFieldDef(CurrentRow);
end;

procedure TSqlitePassCustomFieldDefsDlg.BtOkClick(Sender: TObject);
var
i: integer;
NewCustomFieldDef: TSqlitePassCustomFieldDef;
begin
FDatabaseTypeOptions.CustomFieldDefs.ClearAndFreeItems;
For i := 0 to Pred(CbTableNameComponents.Count) do
    begin
    if (TComboBox(CbTableNameComponents[i]).Text <> '') and (TComboBox(CbFieldNameComponents[i]).Text <> '') then
       begin
       NewCustomFieldDef := TSqlitePassCustomFieldDef.Create;
       With NewCustomFieldDef do
            begin
            TableName := TComboBox(CbTableNameComponents[i]).Text;
            FieldName := TComboBox(CbFieldNameComponents[i]).Text;
            FieldType := StringToFieldType(TComboBox(CbFieldTypeComponents[i]).Text);
            FieldSize := StrToInt(TEdit(EdFieldSizeComponents[i]).Text);
            FieldPrecision := StrToInt(TEdit(EdFieldPrecisionComponents[i]).Text);
            end;
       FDatabaseTypeOptions.CustomFieldDefs.Add(NewCustomFieldDef);
       end;
    end;
end;

{ Events }

procedure TSqlitePassCustomFieldDefsDlg.FormShow(Sender: TObject);
begin
TComboBox(CbTableNameComponents[0]).SetFocus;
end;

procedure TSqlitePassCustomFieldDefsDlg.UpdateFieldNames(CurrentCbFieldName: TComboBox; TableDef: TSqlitePassTableDef);
var
i: integer;
begin
CurrentCbFieldName.Clear;
if Assigned(TableDef)
   then for i := 0 to Pred(TableDef.FieldDefs.Count)
            do CurrentCbFieldName.Items.Add(TableDef.FieldDefs[i].FieldName);
end;

procedure TSqlitePassCustomFieldDefsDlg.OnCbTableNameChange(Sender: TObject);
var
i: integer;
CurrentCbTableName,
CurrentCbFieldName,
LastCbTableName: TComboBox;
begin
CurrentCbTableName := TComboBox(Sender);

CurrentRow := -1;
For i := 0 to Pred(CbTableNameComponents.Count) do
    if CurrentCbTableName = CbTableNameComponents[i] then
       begin
       CurrentRow := i;
       Break;
       end;

CurrentCbFieldName := CbFieldNameComponents[CurrentRow];
UpdateFieldNames(CurrentCbFieldName, Db.TableDefs.TableByName(CurrentCbTableName.Text));

LastCbTableName := CbTableNameComponents[Pred(CbTableNameComponents.Count)];
if CurrentCbTableName = CbTableNameComponents[Pred(CbTableNameComponents.Count)]
   then if Length(CurrentCbTableName.Text) > 0
           then AddNewCustomFieldDef;

if CbTableNameComponents.Count > 1
   then if CurrentCbTableName = CbTableNameComponents[CbTableNameComponents.Count-2]
           then if (Length(CurrentCbTableName.Text) = 0) and (Length(LastCbTableName.Text) = 0)
                    then DeleteCustomFieldDef(Pred(CbTableNameComponents.Count));

end;

procedure TSqlitePassCustomFieldDefsDlg.BtnLoadFromDatabaseClick(
  Sender: TObject);
begin
BtClearAllClick(Sender);
FDatabaseTypeOptions.LoadFromDatabase([soCustomFieldDefs]);
DisplayCustomFieldDefs;
end;

procedure TSqlitePassCustomFieldDefsDlg.BtnSaveToDatabaseClick(Sender: TObject);
begin
BtOkClick(Sender);
FDatabaseTypeOptions.SaveToDatabase([soCustomFieldDefs]);
end;

procedure TSqlitePassCustomFieldDefsDlg.BtClearAllClick(Sender: TObject);
begin
While CbTableNameComponents.Count > 0
      do DeleteCustomFieldDef(Pred(CbTableNameComponents.Count));
FDatabaseTypeOptions.CustomFieldDefs.ClearAndFreeItems;
end;

initialization
 {$IFDEF FPC}
  {$I SqlitePassCustomFieldDefsDialog.lrs}
 {$ENDIF}
end.
 