{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Author : Luc DAVID Email: luckylazarus@free.fr
    Copyright (C) 2007
    Last update : 18/07/2007

  --------------------------------------------------------------------------- }

unit SqlitePassDataTypesDialog;
{$IFDEF FPC}
 {$mode DELPHI}
{$ENDIF}

interface

uses
 {$IFDEF FPC}
  LResources,
 {$ENDIF}
 {$IFDEF WIN32}
  Windows,
 {$ENDIF}
  Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Math, Buttons, ExtCtrls, DB, ComCtrls, SqlitePassDbo, SqlitePassUtils;

type
  TSqlitePassDatatypesDlg = class(TForm)
    PanelBottom: TPanel;
    PanelDatatypes: TPanel;
    BtOk: TButton;
    BtnResetAll: TButton;
    RulesScrollBox: TScrollBox;
    HeaderControl1: THeaderControl;
    SbDelete: TSpeedButton;
    SbAdd: TSpeedButton;
    BtDefault: TButton;
    BtnSaveToDatabase: TButton;
    BtnLoadFromDatabase: TButton;
    Button1: TButton;
    procedure SbAddClick(Sender: TObject);
    procedure SbDeleteClick(Sender: TObject);
    procedure BtOkClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure BtnResetAllClick(Sender: TObject);
    procedure BtDefaultClick(Sender: TObject);
    procedure BtnLoadFromDatabaseClick(Sender: TObject);
    procedure BtnSaveToDatabaseClick(Sender: TObject);
  private
    RowTop: Integer;
    CbMappingModeComponents: TList;
    CbDataTypeComponents: TList;
    EdDataTypeNameComponents: Tlist;
    CurrentRow: Integer;
    FDatabaseTypeOptions: TSqlitePassDatabaseDataTypeOptions;
    FDatabaseType: TSqlitePassDatabaseType;
    Procedure AddNewDataTypeRule;
    Procedure DeleteDataTypeRule(Index: Integer);
    procedure DisplayConversionRules;
    Procedure OnEdDataTypeNameChange(Sender: TObject);
    Procedure OnEdDataTypeNameEnter(Sender: TObject);
    Procedure OnCbMappingModeEnter(Sender: TObject);
    Procedure OnCbDataTypeEnter(Sender: TObject);
  public
    constructor Create(AOwner: TComponent; DatabaseTypeOptions: TSqlitePassDatabaseDataTypeOptions; DatabaseType: TSqlitePassDatabaseType); reintroduce;
    destructor Destroy; override;
  end;

var
  SqlitePassDatatypesDlg: TSqlitePassDatatypesDlg;

implementation

{$IFNDEF FPC}
 {$R *.DFM}
{$ENDIF}

{ Constructor and Destructor }

constructor TSqlitePassDatatypesDlg.Create(AOwner: TComponent; DatabaseTypeOptions: TSqlitePassDatabaseDataTypeOptions; DatabaseType: TSqlitePassDatabaseType);
begin
inherited Create(AOwner);
FDatabaseTypeOptions := DatabaseTypeOptions;
FDatabaseType := DatabaseType;
RowTop := -16;
EdDataTypeNameComponents:= TList.Create;
CbMappingModeComponents:= TList.Create;
CbDataTypeComponents:= TList.Create;
DisplayConversionRules;
end;

destructor TSqlitePassDatatypesDlg.Destroy;
begin
CbMappingModeComponents.Free;
CbDataTypeComponents.Free;
EdDataTypeNameComponents.Free;
inherited;
end;

procedure TSqlitePassDatatypesDlg.DisplayConversionRules;
var
i: Integer;
begin
Try
  RulesScrollBox.Visible := False;
  For i := 0 to Pred(FDatabaseTypeOptions.TranslationRules.Count) do
      begin
      AddNewDataTypeRule;
      TEdit(EdDataTypeNameComponents[i]).Text := FDatabaseTypeOptions.TranslationRules[i].DataTypeName;
      TComboBox(CbMappingModeComponents[i]).ItemIndex := Ord(FDatabaseTypeOptions.TranslationRules[i].MappingMode);
      TComboBox(CbDataTypeComponents[i]).ItemIndex := TComboBox(CbDataTypeComponents[i]).Items.IndexOf(FieldTypeToString(FDatabaseTypeOptions.TranslationRules[i].FieldType));
      end;
  if EdDataTypeNameComponents.Count = 0
     then AddNewDataTypeRule;
Finally
  RulesScrollBox.Visible := True;
end;
end;

procedure TSqlitePassDatatypesDlg.AddNewDataTypeRule;
var
NewCbMappingMode: TComboBox;
NewCbDataType: TComboBox;
NewEdDataTypeName: TEdit;
ft: TFieldType;
begin
Inc(RowTop, 20);
NewEdDataTypeName := TEdit.Create(RulesScrollBox);
With NewEdDataTypeName do
     begin
     Parent := RulesScrollBox;
     Text := '';
     Top  := RowTop;
     Left := 3;
     Width := 160;
     Height := 20;
     OnChange := OnEdDataTypeNameChange;
     OnEnter := OnEdDataTypeNameEnter;
     end;

NewCbMappingMode := TComboBox.Create(RulesScrollBox);
With NewCbMappingMode do
     begin
     Parent := RulesScrollBox;
     Top  := RowTop;
     Left := 171;
     Width := 155;
     Height := 20;
     Items.Add('mmExact');
     Items.Add('mmPartial');
     Items.Add('mmAll');
     Style := csDropDownList;
     ItemIndex := 2;
     OnEnter := OnCbMappingModeEnter;
     end;

NewCbDataType := TComboBox.Create(RulesScrollBox);
With NewCbDataType do
     begin
     Parent := RulesScrollBox;
     Top  := RowTop;
     Left := 332;
     Width := 155;
     Height := 20;
     For ft := Low(TFieldType) to High(TFieldType)
         do  Items.Add(FieldTypeToString(ft));     Style := csDropDownList;
     Sorted := True;
     ItemIndex := 0;
     OnEnter := OnCbDataTypeEnter;
     end;
EdDataTypeNameComponents.Add(NewEdDataTypeName);
CbMappingModeComponents.Add(NewCbMappingMode);
CbDataTypeComponents.Add(NewCbDataType);
end;


procedure TSqlitePassDatatypesDlg.DeleteDataTypeRule(Index: Integer);
begin
TEdit(EdDataTypeNameComponents[Index]).Free;
TComboBox(CbMappingModeComponents[Index]).Free;
TComboBox(CbDataTypeComponents[Index]).Free;
EdDataTypeNameComponents.Delete(Index);
CbMappingModeComponents.Delete(Index);
CbDataTypeComponents.Delete(Index);
Dec(RowTop,20);
end;

{ Buttons Events }

procedure TSqlitePassDatatypesDlg.SbAddClick(Sender: TObject);
begin
AddNewDataTypeRule;
end;

procedure TSqlitePassDatatypesDlg.SbDeleteClick(Sender: TObject);
begin
DeleteDataTypeRule(CurrentRow);
end;

procedure TSqlitePassDatatypesDlg.BtOkClick(Sender: TObject);
var
i: integer;
NewConversionRule: TSqlitePassFieldTypeTranslationRule;
begin
FDatabaseTypeOptions.TranslationRules.ClearAndFreeItems;
For i := 0 to Pred(EdDataTypeNameComponents.Count) do
    begin
    if TEdit(EdDataTypeNameComponents[i]).Text <> '' then
       begin
       NewConversionRule := TSqlitePassFieldTypeTranslationRule.Create;
       With NewConversionRule do
            begin
            DataTypeName := TEdit(EdDataTypeNameComponents[i]).Text;
            MappingMode := TSqlitePassDataTypeMappingMode(TComboBox(CbMappingModeComponents[i]).ItemIndex);
            FieldType := StringToFieldType(TComboBox(CbDataTypeComponents[i]).Text);
            end;
       FDatabaseTypeOptions.TranslationRules.Add(NewConversionRule);
       end;
    end;
end;

{ Events }

procedure TSqlitePassDatatypesDlg.FormShow(Sender: TObject);
begin
TEdit(EdDataTypeNameComponents[0]).SetFocus;
end;

procedure TSqlitePassDatatypesDlg.OnEdDataTypeNameEnter(Sender: TObject);
var
i: integer;
begin
CurrentRow := -1;
For i := 0 to Pred(EdDataTypeNameComponents.Count) do
    if TEdit(Sender) = EdDataTypeNameComponents[i] then
       begin
       CurrentRow := i;
       Break;
       end;
end;

procedure TSqlitePassDatatypesDlg.OnEdDataTypeNameChange(Sender: TObject);
var
CurrentEdDataTypeName,
LastEdDataTypeName: TEdit;
begin
CurrentEdDataTypeName := TEdit(Sender);
LastEdDataTypeName := EdDataTypeNameComponents[Pred(EdDataTypeNameComponents.Count)];
if CurrentEdDataTypeName = EdDataTypeNameComponents[Pred(EdDataTypeNameComponents.Count)]
   then if Length(CurrentEdDataTypeName.Text) = 1
           then AddNewDataTypeRule;
if EdDataTypeNameComponents.Count > 1
   then if CurrentEdDataTypeName = EdDataTypeNameComponents[EdDataTypeNameComponents.Count-2]
           then if (Length(CurrentEdDataTypeName.Text) = 0) and (Length(LastEdDataTypeName.Text) = 0)
                    then DeleteDataTypeRule(Pred(EdDataTypeNameComponents.Count));
end;

procedure TSqlitePassDatatypesDlg.OnCbMappingModeEnter(Sender: TObject);
var
i: integer;
begin
CurrentRow := -1;
For i := 0 to Pred(CbMappingModeComponents.Count) do
    if TEdit(Sender) = CbMappingModeComponents[i] then
       begin
       CurrentRow := i;
       Break;
       end;
end;

procedure TSqlitePassDatatypesDlg.OnCbDataTypeEnter(Sender: TObject);
var
i: integer;
begin
CurrentRow := -1;
For i := 0 to Pred(CbDataTypeComponents.Count) do
    if TEdit(Sender) = CbDataTypeComponents[i] then
       begin
       CurrentRow := i;
       Break;
       end;
end;

procedure TSqlitePassDatatypesDlg.BtnResetAllClick(Sender: TObject);
begin
While EdDataTypeNameComponents.Count > 0
      do DeleteDataTypeRule(Pred(EdDataTypeNameComponents.Count));
DisplayConversionRules;
end;

procedure TSqlitePassDatatypesDlg.BtDefaultClick(Sender: TObject);
begin
While EdDataTypeNameComponents.Count > 0
      do DeleteDataTypeRule(Pred(EdDataTypeNameComponents.Count));
FDatabaseTypeOptions.TranslationRules.SetDefaultTranslationRules(FDatabaseType);
DisplayConversionRules;
end;

procedure TSqlitePassDatatypesDlg.BtnLoadFromDatabaseClick(
  Sender: TObject);
begin
While EdDataTypeNameComponents.Count > 0
      do DeleteDataTypeRule(Pred(EdDataTypeNameComponents.Count));
FDatabaseTypeOptions.LoadFromDatabase([soTranslationRules]);
DisplayConversionRules;
end;

procedure TSqlitePassDatatypesDlg.BtnSaveToDatabaseClick(Sender: TObject);
begin
BtOkClick(Sender);
FDatabaseTypeOptions.SaveToDatabase([soTranslationRules]);
end;

initialization
 {$IFDEF FPC}
  {$I SqlitePassDataTypesDialog.lrs}
 {$ENDIF}
end.
 