{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Author : Luc DAVID Email: luckylazarus@free.fr
    Copyright (C) 2006 - 2008

    Major changes are indicated in the \Documentation\Changes.pdf file
    Last update 13.01.2008

  --------------------------------------------------------------------------- }


unit SqlitePassDbo;
{$i SqlitePassDbo.inc}

{ Turns on/off debug output to TMemo }
{.$DEFINE DEBUG}
{.$DEFINE DEBUG_SQLStmt}

Interface
uses
{$IFNDEF FPC}
DBCommon,
{$ENDIF}
Classes, {$IFNDEF linux}Windows,{$ENDIF} SysUtils, Db, Forms, Math,
{$IFDEF DEBUG or $IDDEF DEBUG_SQLStmt}
StdCtrls, { To use a TMemo as dirty debug screen }
{$ENDIF}
Dialogs, { To be removed - debug }
SqlitePassUtils, SqlitePassApi_v3, SqlitePassConst, SqlitePassErrorLang, SqlitePassEngine,
SqlitePassKexiDef;


Const
SqlitePassBlobFieldTypes = [ftBytes, ftVarBytes, ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftTypedBinary];

Type
 { forward declarations }
 TSqlitePassDatabase = Class;
 TSqlitePassDataset = class;
 TSqlitePassDatasets = class;
 TSqlitePassTableDef = class;
 TSqlitePassTableDefs = class;
 TSqlitePassDatabaseIndexDefs = class;
 TSqlitePassQueryDefs = class;
 TSqlitePassViews = class;
 TSqlitePassTriggers = class;
 TSqlitePassTableFieldDef = class;
 TSqlitePassTableFieldDefs = class;
 TSqlitePassIndexColumns = class;
 TSqlitePassIndex = class;
 TSqlitePassDatasetIndex = class;
 TSqlitePassTranslator = class;
 TSqlitePassDatabasesAttached = class;
 TSqlitePassSelectStmtFieldDefs = class;
 TSqlitePassSelectStmtTableDef = class;
 TSqlitePassRecordset = class;
 TSqlitePassMemRecord = class;
 TSqlitePassSelectStmt = class;
 TSqlitePassSqlStmt = class;
 TSqlitePassDatabaseDataTypeOptions = class;
 
 TSqlitePassDatabaseObjectTypes =(dboAll, dboTableDefs, dboQueryDefs, dboIndexDefs, dboViews, dboTriggers);
 TSqlitePassDatabaseType = (dbtSqlitePass, dbtKexi, dbtSqliteAdmin, dbtFpc, dbtSqliteExpert, dbtUnknown);
 TSqlitePassDatabaseState = (dbsInternalOpen, dbsActive, dbsInactive);
 TSqlitePassPostType = (ptInsert, ptUpdate, ptDelete);
 TSqlitePassQuoteStyle = (qsSingleQuote, qsDoubleQuote, qsBracket, qsNone);


 { TSqlitePassDatabase Events }
 TConnectEvent = procedure(Database: TSqlitePassDatabase) of object;
 TDataTypeConversion = procedure (Database: TSqlitePassDatabase; TableDef: TSqlitePassTableDef;
                                  DataTypeName: String; var DataType: TFieldType) of object;



 { TVersionInfo }
 TSqlitePassDatabaseVersionInfo = Class(TPersistent)
     Private
         FComponent   : String;
         FDatabase    : TSqlitePassDatabase;
         FPackage     : String;
         FKexi        : String;
         FSchema      : Integer;
         FSqliteLib   : String;
         FUserTag     : Integer;
         function GetFSchema: Integer;
         function GetFSqliteEngine: String;
         function GetFUserTag: Integer;
         procedure SetFKexi(const AValue: String);
         procedure SetFPackage(const AValue: String);
         procedure SetFSchema(const AValue: Integer);
         procedure SetFSqliteEngine(const AValue: String);
         procedure SetFUserTag(const AValue: Integer);
         procedure SetFComponent(const Value: String);
     Public
         constructor Create(Database: TSqlitePassDatabase);
         destructor Destroy; override;
     Published
         Property Component     : String Read FComponent        Write SetFComponent;
         Property Database      : Integer Read GetFSchema       Write SetFSchema;
         Property Kexi          : String  Read FKexi            Write SetFKexi;
         Property Package       : String  Read FPackage         Write SetFPackage;
         Property SqliteLibrary : String  Read GetFSqliteEngine Write SetFSqliteEngine;
         Property UserTag       : Integer Read GetFUserTag      Write SetFUserTag;
     end;

TSqlitePassSynchronous = (Off, Normal, Full);
TSqlitePassTempStore = (Default, ToFile, ToMemory);
TSqlitePassEncoding = (UTF_8, UTF_16, UTF_16le, UTF_16be);
TSqlitePassPageSize = 0..8192;//512, 1024, 2048, 4096, 8192

 { TSqlitePassDatabaseOptions }
 TSqlitePassDatabaseOptions = Class(TPersistent)
    Private
         FDatabase: TSqlitePassDatabase;
         FAutoVacuum: Boolean;
         FCacheSize: Integer;
         FDefaultCacheSize: Integer;
         FCaseSensitiveLike: Boolean;
         FCountChanges: Boolean;
         FEmptyResultCallback: Boolean;
         FEncoding: String;
         FFullColumnNames: Boolean;
         FQuoteStyle: TSqlitePassQuoteStyle;
         FQuoteOpen: Char;
         FQuoteClose: Char;
         FShortColumnNames: Boolean;
         FPageSize: TSqlitePassPageSize;
         FSynchronous : TSqlitePassSynchronous;
         FTemporaryStorage: TSqlitePassTempStore;
         function GetFAutoVacuum: Boolean;
         function GetFCacheSize: Integer;
         function GetFCountChanges: Boolean;
         function GetFDefaultCacheSize: Integer;
         function GetFEmptyResultCallBack: Boolean;
         function GetFEncoding: TSqlitePassEncoding;
         function GetFFullColumnNames: Boolean;
         function GetFPageSize: TSqlitePassPageSize;
         function GetFSynchronous: TSqlitePassSynchronous;
         function GetFTemporaryStorage: TSqlitePassTempStore;
         procedure SetFAutoVacuum(const AValue: Boolean);
         procedure SetFCacheSize(const AValue: Integer);
         procedure SetFCaseSensitiveLike(const AValue: Boolean);
         procedure SetFCountChanges(const AValue: Boolean);
         procedure SetFDefaultCacheSize(const AValue: Integer);
         procedure SetFEmptyResultCallBack(const AValue: Boolean);
         procedure SetFEncoding(const AValue: TSqlitePassEncoding);
         procedure SetFFullColumnNames(const AValue: Boolean);
         procedure SetFPageSize(AValue: TSqlitePassPageSize);
         procedure SetFQuoteStyle(const AValue: TSqlitePassQuoteStyle);
         procedure SetFSynchronous(const AValue: TSqlitePassSynchronous);
         procedure SetFTemporaryStorage(const AValue: TSqlitePassTempStore);
    Public
         constructor Create(Database: TSqlitePassDatabase);
         destructor Destroy; override;
    Published
         Property AutoVacuum: Boolean Read GetFAutoVacuum Write SetFAutoVacuum;
         Property CacheSize: Integer Read GetFCacheSize Write SetFCacheSize;
         Property CaseSensitiveLike: Boolean Read FCaseSensitiveLike Write SetFCaseSensitiveLike;
         Property CountChanges: Boolean Read GetFCountChanges Write SetFCountChanges;
         Property DefaultCacheSize: Integer Read GetFDefaultCacheSize Write SetFDefaultCacheSize;
         Property EmptyResultCallback: Boolean Read GetFEmptyResultCallBack Write SetFEmptyResultCallBack;
         Property Encoding: TSqlitePassEncoding Read GetFEncoding Write SetFEncoding;
         Property FullColumnNames: Boolean Read GetFFullColumnNames Write SetFFullColumnNames; //Think to Set Also ShortColumnNames
         Property PageSize: TSqlitePassPageSize Read GetFPageSize Write SetFPageSize;
         Property QuoteStyle: TSqlitePassQuoteStyle Read FQuoteStyle Write SetFQuoteStyle;
         Property Synchronous : TSqlitePassSynchronous Read GetFSynchronous Write SetFSynchronous;
         Property TemporaryStorage: TSqlitePassTempStore Read GetFTemporaryStorage Write SetFTemporaryStorage;
    end;

 { Datatypes detection and conversion }

 TSqlitePassDateStorage = (dsDateTime, dsDbSpecific, dsJulian, dsText, dsUnix);
 TSqlitePassDataTypeDetectionMode = (dmTypeName, dmDbSpecific, dmCustom, dmCustomFieldDefs, dmForceStr, dmNone);
 TSqlitePassDataTypeMappingMode = (mmExact, mmExactNoCase, mmPartial, mmPartialNoCase, mmAll);
 TSqlitePassDataTypeStorageOptions = set of (soProperties, soTranslationRules, soCustomFieldDefs, soManual);

 { Database FieldTypes mapping to Pascal FieldTypes }
 { TSqlitePassFieldTypeTranslationRule }
 TSqlitePassFieldTypeTranslationRule = Class
 Private
 FDatatypeName: String;
 FMappingMode: TSqlitePassDataTypeMappingMode;
 FFieldType: TFieldType;
 Public
 Property DatatypeName: String Read FDatatypeName Write FDatatypeName;
 Property MappingMode: TSqlitePassDataTypeMappingMode Read FMappingMode Write FMappingMode;
 Property FieldType: TFieldType Read FFieldType Write FFieldType;
 end;

 { TSqlitePassFieldTypesTranslationRules }
 TSqlitePassFieldTypesTranslationRules = Class(TSqlitePassObjectList)
  Private
    FDataTypeOptions: TSqlitePassDatabaseDataTypeOptions;
  Protected
    Function GetItem(Index: Integer): TSqlitePassFieldTypeTranslationRule;
    Procedure SetItem(Index: Integer; FieldTypeTranslationRule: TSqlitePassFieldTypeTranslationRule);
  Public
    constructor Create(DataTypeOptions: TSqlitePassDatabaseDataTypeOptions);
    Destructor Destroy; override;
    Function Add(FieldTypeTranslator: TSqlitePassFieldTypeTranslationRule): Integer;
    procedure ClearAndFreeItems;
    Procedure SetDefaultTranslationRules(DatabaseType: TSqlitePassDatabaseType);
    Function ConvertNativeFieldTypeToPascalFieldType(FieldTypeName: String): TFieldType;
    Property Items[Index: Integer]: TSqlitePassFieldTypeTranslationRule read GetItem write SetItem; default;
  end;

 { TSqlitePassCustomFiedlDef }
 TSqlitePassCustomFieldDef = Class
 Private
 FFieldName: String;
 FFieldType: TFieldType;
 FFieldSize: Integer;
 FFieldPrecision: Integer;
 FTableName: String;
 Public
 Property TableName: String Read FTableName Write FTableName;
 Property FieldName: String Read FFieldName Write

 FFieldName;
 Property FieldType: TFieldType Read FFieldType Write FFieldType;
 Property FieldSize: Integer Read FFieldSize Write FFieldSize;
 Property FieldPrecision: Integer Read FFieldPrecision Write FFieldPrecision;
 end;

 { TSqlitePassCustomFieldDefs }
 TSqlitePassCustomFieldDefs = Class(TSqlitePassObjectList)
  Protected
    Function GetItem(Index: Integer): TSqlitePassCustomFieldDef;
    Procedure SetItem(Index: Integer; CustomFieldDef: TSqlitePassCustomFieldDef);
  Public
    Constructor Create;
    Destructor Destroy; override;
    Function Add(CustomFieldDef: TSqlitePassCustomFieldDef): Integer;
    procedure ClearAndFreeItems;
    Property Items[Index: Integer]: TSqlitePassCustomFieldDef read GetItem write SetItem; default;
  end;


 TSqlitePassDatabaseDataTypeOptions = Class(TPersistent)
      Private
         FBooleanExtension     : String;
         FCustomFieldDefs      : TSqlitePassCustomFieldDefs;
         FDatabase             : TSqlitePassDatabase;
         FDetectionMode        : TSqlitePassDataTypeDetectionMode;
         FDateSeparator        : Char;
         FDateStorage          : TSqlitePassDateStorage;
         FDecimalSeparator     : Char;
         FDefaultFieldType     : TFieldType;
         FFieldTypesTranslationRules : TSqlitePassFieldTypesTranslationRules;
         FLoadOptions: TSqlitePassDataTypeStorageOptions;
         FLongTimeFormat       : String;
         FSaveOptions          : TSqlitePassDataTypeStorageOptions;
         FShortDateFormat      : String;
         FTimeSeparator        : Char;
         procedure ReadTranslationRules(Reader: TReader);
         procedure WriteTranslationRules(Writer: TWriter);
         procedure ReadCustomFieldDefs(Reader: TReader);
         procedure WriteCustomFieldDefs(Writer: TWriter);
        Protected
         procedure DefineProperties(Filer: TFiler); override;
        Public
          { ----- Constructor / Destructor ----- }
         constructor Create(Database: TSqlitePassDatabase);
         destructor Destroy; override;
         { --- }
         Procedure ApplyCustomFieldDefs;
         function  GetOwner: TPersistent; override;
         procedure SetDefaultValues(DatabaseType: TSqlitePassDatabaseType; SetTranslationRules: Boolean=false);
         Procedure LoadFromDatabase(_LoadOptions: TSqlitePassDataTypeStorageOptions = []);
         Procedure SaveToDatabase(_SaveOptions: TSqlitePassDataTypeStorageOptions = []);
       Published
         { ----- Properties ----- }
         Property BooleanExtension     : String Read FBooleanExtension Write FBooleanExtension;
         Property CustomFieldDefs      : TSqlitePassCustomFieldDefs Read FCustomFieldDefs Write FCustomFieldDefs;
         Property DefaultFieldType     : TFieldType Read FDefaultFieldType Write FDefaultFieldType;
         Property DetectionMode        : TSqlitePassDataTypeDetectionMode Read FDetectionMode Write FDetectionMode;
         Property DateSeparator        : Char Read FDateSeparator Write FDateSeparator;
         Property DateStorage          : TSqlitePassDateStorage Read FDateStorage Write FDateStorage;
         Property DecimalSeparator     : Char Read FDecimalSeparator Write FDecimalSeparator;
         Property LoadOptions          : TSqlitePassDataTypeStorageOptions Read FLoadOptions Write FLoadOptions;
         Property LongTimeFormat       : String Read FLongTimeFormat Write FLongTimeFormat;
         Property ShortDateFormat      : String Read FShortDateFormat Write FShortDateFormat;
         Property SaveOptions          : TSqlitePassDataTypeStorageOptions Read FSaveOptions Write FSaveOptions;
         Property TimeSeparator        : Char Read FTimeSeparator Write FTimeSeparator;
         Property TranslationRules     : TSqlitePassFieldTypesTranslationRules Read FFieldTypesTranslationRules Write FFieldTypesTranslationRules;
      end;


 { TSqlitePassDatabase }

 TSqlitePassDatabase = Class(TComponent)
       Private
         FActive             : Boolean;
         FCollatingOrder     : String;
         FDatabase           : String;
         FDatasets           : TSqlitePassDatasets;
         FTableDefs          : TSqlitePassTableDefs;
         FIndexDefs          : TSqlitePassDatabaseIndexDefs;
         FQueryDefs          : TSqlitePassQueryDefs;
         FViews              : TSqlitePassViews;
         FTriggers           : TSqlitePassTriggers;
         FDatabases          : TSqlitePassDatabasesAttached;
         FDatabaseType       : TSqlitePassDatabaseType;
         FDataTypeOptions    : TSqlitePassDatabaseDataTypeOptions;
         FEngine             : TSqlitePassEngine;
         FLongTimeFormat     : String;
         FTranslator         : TSqlitePassTranslator;
         FOptions            : TSqlitePassDatabaseOptions;
         FQueryTimeout       : Integer;
         FReadOnly           : Boolean;
         FShowSysObjects     : Boolean;
         FSmartOpen          : Boolean;
         FState              : TSqlitePassDatabaseState;
         FUser               : Integer;
         FVersionInfo        : TSqlitePassDatabaseVersionInfo;
         FSQLiteLibrary      : String;

         { ----- Events ---- }
         FAfterConnect       : TConnectEvent;
         FAfterDisconnect    : TConnectEvent;
         FBeforeConnect      : TConnectEvent;
         FBeforeDisconnect   : TConnectEvent;
         FDataTypeConversion : TDataTypeConversion;

         { Private Functions and Procedures }
         function CheckCanOpen: Boolean;
         Procedure SetFDatabase(Value : String);
         procedure SetFDatabaseType(const AValue: TSqlitePassDatabaseType);
         Procedure SetFDatabaseVersion(Value : String);
         procedure SetFUser(const AValue: Integer);
         Procedure SetFVersionInfo(Value : String);
         Function  GetFCollatingOrder:String;
         Procedure SetFShowSysObjects(Value : Boolean);
         Procedure SetFSQLiteLibrary(Value : String);
         Procedure SetFReadOnly(Value : Boolean);
         function GetFActive: Boolean;
         Procedure SetFActive(Value : Boolean);
         function GetFEngine: TSqlitePassEngine;
         procedure ClearDefinitions;
         { ------ Pragma ------ }
         procedure setPragma(PragmaName:string; value:Boolean);overload;
         procedure setPragma(PragmaName:string; value:integer);overload;
         procedure setPragma(PragmaName, value:string);overload;
         function GetPragma(PragmaName:string):integer;
       Protected
         Procedure Loaded; override;
       Public
         { ----- Constructor / Destructor ----- }
         Constructor Create(AOwner : TComponent); override;
         Destructor Destroy; override;
         { ----- Procedures / Functions ----- }
         Procedure Idle;
         Procedure Open;
         Procedure Close;
         Function Compact: Integer;
         Function AttachDatabase(FAttachedDatabase, DatabaseAlias: String): Boolean;
         Function DetachDatabase(DatabaseAlias: String): Boolean;
         procedure RefreshDefinitions;
         Function IsSystemTable(TableName: String): Boolean;
         { ----- Properties ----- }
         Property Engine: TSqlitePassEngine Read GetFEngine;
         Property Translator: TSqlitePassTranslator Read FTranslator;
         property Databases: TSqlitePassDatabasesAttached Read FDatabases;
         Property Datasets: TSqlitePassDatasets Read FDatasets Write FDatasets;
         Property IndexDefs: TSqlitePassDatabaseIndexDefs  Read FIndexDefs Write FIndexDefs;
         property QueryDefs: TSqlitePassQueryDefs Read FQueryDefs Write FQueryDefs;
         property TableDefs: TSqlitePassTableDefs Read FTableDefs Write FTableDefs;
         property Views: TSqlitePassViews Read FViews Write FViews;
         property Triggers: TSqlitePassTriggers Read FTriggers Write FTriggers;
       Published
         { ----- Properties ----- }
         Property Connected            : Boolean Read GetFActive           Write SetFActive Default False;
         Property Database             : String  Read FDatabase            Write SetFDatabase;
         Property DatabaseType         : TSqlitePassDatabaseType Read FDatabaseType Write FDatabaseType;
         Property DatatypeOptions      : TSqlitePassDatabaseDataTypeOptions Read FDatatypeOptions Write FDatatypeOptions;
         Property Options              : TSqlitePassDatabaseOptions Read FOptions Write FOptions;
         Property QueryTimeout         : Integer Read FQueryTimeout        Write FQueryTimeout;
         Property ReadOnly             : Boolean Read FReadOnly            Write SetFReadOnly default False;
         Property ShowSystemObjects    : Boolean Read FShowSysObjects      Write SetFShowSysObjects;
         Property SmartOpen            : Boolean Read FSmartOpen           Write FSmartOpen;
         Property SQLiteLibrary        : String  Read FSQLiteLibrary       Write SetFSQLiteLibrary;
         Property UserTag              : Integer Read FUser                Write SetFUser;
         Property VersionInfo          : TSqlitePassDatabaseVersionInfo  Read FVersionInfo   Write FVersionInfo;
          { ----- Events ----- }
         Property OnAfterConnect       : TConnectEvent Read FAfterConnect        Write FAfterConnect;
         Property OnAfterDisconnect    : TConnectEvent Read FAfterDisconnect     Write FAfterDisconnect;
         Property OnBeforeConnect      : TConnectEvent Read FBeforeConnect       Write FBeforeConnect;
         Property OnBeforeDisconnect   : TConnectEvent Read FBeforeDisconnect    Write FBeforeDisconnect;
         Property OnDataTypeConversion : TDataTypeConversion Read FDataTypeConversion Write FDataTypeConversion;
      end;

TSqlitePassGenericFieldDef = Class
  Private
   FNoConstraints: Boolean;
   FAutoInc: Boolean;
   FUnique: Boolean;
   FPrimaryKey: Boolean;
   FInternalPrimaryKey: Boolean;
   FForeignKey: Boolean;
   FNotNull: Boolean;
   FNotEmpty: Boolean; { only legal for string-like and blob fields }
   FIndexed: Boolean;
   FDatabaseOrigineNameIndex: Integer;
   FTableOrigineNameIndex: Integer;


   FFieldName: String;
   FFieldFullName: String;
   FDisplayName: String;
   FSize: Word;
   FPrecision: Integer;
   FNativeDataType: String;
   FNativeDataTypeCode: Integer;
   FDataType: TFieldType;
   FDefaultValue: String;
   FFieldNo: Integer;
   FHint: String;
  Public
   Property AutoInc: Boolean Read FAutoInc;
   { write is enabled to force a DataType conversion }
   Property DataType: TFieldType Read FDataType Write FDataType;
   Property DefaultValue: String Read FDefaultValue;
   { write is enabled to enable changing default display name }
   Property DisplayName: String Read FDisplayName Write FDisplayName;
   Property FieldName: String Read FFieldName;
   Property FieldNo: Integer Read FFieldNo;
   Property ForeignKey: Boolean Read FForeignKey;
   { write is enabled to enable changing default hint message }
   Property Hint: String Read FHint Write FHint;
   Property Indexed: Boolean Read FIndexed;
   Property InternalPrimaryKey: Boolean Read FInternalPrimaryKey;
   Property NativeDataType: String Read FNativeDataType;
   Property NativeDataTypeCode: Integer Read FNativeDataTypeCode Write FNativeDataTypeCode;
   Property NotEmpty: Boolean Read FNotEmpty;
   Property NoConstraints: Boolean Read FNoConstraints;
   Property NotNull: Boolean Read FNotNull;
   Property Precision: Integer Read FPrecision;
   Property PrimaryKey: Boolean Read FPrimaryKey;
   Property Size: Word Read FSize;
   Property Unique: Boolean Read FUnique;
  end;

TSqlitePassTableFieldDefs = class(TSqlitePassObjectList)
  private
  FTableDef: TSqlitePassTableDef;
  Protected
  procedure ClearAndFreeItems;
  function GetItem(Index: Integer): TSqlitePassTableFieldDef;
  procedure SetItem(Index: Integer; const Value: TSqlitePassTableFieldDef);
  public
  Constructor Create(Owner: TSqlitePassTableDef);
  Destructor Destroy; override;
  function FieldDefByName(const Value: string): TSqlitePassTableFieldDef;
  function FindFieldDef(const Value: string): TSqlitePassTableFieldDef;
  property Count;
  property Items[Index: Integer]: TSqlitePassTableFieldDef read GetItem write SetItem; default;
  property TableDef: TSqlitePassTableDef read FTableDef;
  end;

TSqlitePassTableFieldDef = Class(TSqlitePassGenericFieldDef)
  Private
  FFieldDefs: TSqlitePassTableFieldDefs;
  function GetFFieldFullName: String;
  Public
  Constructor Create(Owner: TSqlitePassTableFieldDefs);
  Destructor Destroy; override;
  Property FieldFullName: String Read GetFFieldFullName;
  Property FieldDefs: TSqlitePassTableFieldDefs Read FFieldDefs;
  end;

TSqlitePassSelectStmtFieldDef = Class(TSqlitePassGenericFieldDef)
  Private
  FFieldDefs: TSqlitePassSelectStmtFieldDefs;
  Public
  Constructor Create(Owner: TSqlitePassSelectStmtFieldDefs);
  Destructor Destroy; override;
  procedure Assign(FieldDef: TSqlitePassTableFieldDef);
  Property FieldDefs: TSqlitePassSelectStmtFieldDefs Read FFieldDefs;
  end;
    
TSqlitePassSelectStmtFieldDefs = class(TSqlitePassObjectList)
  private
  FTableDef: TSqlitePassSelectStmtTableDef;
  function GetItem(Index: Integer): TSqlitePassSelectStmtFieldDef;
  procedure SetItem(Index: Integer; const Value: TSqlitePassSelectStmtFieldDef);
  procedure ClearAndFreeItems;
  public
  Constructor Create(Owner: TSqlitePassSelectStmtTableDef);
  Destructor Destroy; override;
  procedure Assign(FieldDefs: TSqlitePassTableFieldDefs);
  property Count;
  property Items[Index: Integer]: TSqlitePassSelectStmtFieldDef read GetItem write SetItem; default;
  property TableDef: TSqlitePassSelectStmtTableDef read FTableDef;
  end;
  

{ TSqliteAttachedDatabases : Maintain a list of all attached databases }

TSqlitePassDatabaseAttached = Class(TObject)
       Private
         FAliasName          : String;
         FDatabase           : String;
         FDatabasesList      : TSqlitePassDatabasesAttached;
         FEngine             : TSqlitePassEngine;
       Public
         { ----- Constructor / Destructor ----- }
         Constructor Create(Databases: TSqlitePassDatabasesAttached);
         Destructor Destroy; override;
         { ----- Properties ----- }
         Property AliasName: String Read FAliasName;
         Property Database: String  Read FDatabase Write FDatabase;
         Property Engine: TSqlitePassEngine Read FEngine;
      end;

TSqlitePassDatabasesAttached = Class(TSqlitePassObjectList)
  Private
   FDatabase: TSqlitePassDatabase;
  Protected
    Function GetItem(Index: Integer): TSqlitePassDatabaseAttached;
    Procedure SetItem(Index: Integer; Database: TSqlitePassDatabaseAttached);
    procedure DetachAll;
  public
    constructor Create(Owner: TSqlitePassDatabase);
    destructor Destroy; override;
    function DatabaseByName(Const Name: String): TSqlitePassDatabaseAttached;
    function FindDatabase(const Value: string): TSqlitePassDatabaseAttached;
    property Items[Index: Integer]: TSqlitePassDatabaseAttached read GetItem write SetItem; default;
    property Count;
  end;

{ TableDefs }


TSqlitePassSelectStmtTableDefs = class(TSqlitePassObjectList)
  private
  FSQLStmt: TSqlitePassSelectStmt;
  Protected
  procedure ClearAndFreeItems;
  function GetItem(Index: Integer): TSqlitePassSelectStmtTableDef;
  procedure SetItem(Index: Integer; const Value: TSqlitePassSelectStmtTableDef);
  public
  constructor Create(Owner: TSqlitePassSelectStmt);
  destructor Destroy; override;
//  function TableByName(const Value: string): TSqlitePassSelectStmtTableDef;
  function FindTable(const Value: string): TSqlitePassSelectStmtTableDef;
//  function FindFieldDef(TableName, FieldName: String): TSqlitePassFieldDef;
//  function FindFieldTableName(FieldName: String; TableNames: TStringList): String;
  property Count;
  property Items[Index: Integer]: TSqlitePassSelectStmtTableDef read GetItem write SetItem; default;
  end;

TSqlitePassGenericTableDef = class
  private
  FAliasName: String;
  FCanModify: Boolean;
  FDatabaseOrigineNameIndex: Integer;
  FOnConflict: String;
  FPrimaryKeyFieldNo: Integer;
  FSql: String;
  FTableName: String;
  FTableFullName: String;
  FTemporary: Boolean;
  Public
  Property AliasName: String Read FAliasName;
  Property CanModify: Boolean Read FCanModify;
  Property OnConlict: String Read FOnConflict;
  Property PrimaryKeyFieldNo: Integer Read FPrimaryKeyFieldNo;
  property Sql: String Read FSql;
  property TableName: String Read FTableName;
  property TableFullName: String Read FTableFullName;
  Property Temporary: Boolean Read FTemporary;
  end;

TSqlitePassSelectStmtTableDef = class(TSqlitePassGenericTableDef)
  private
  { Owner }
  FTableDefs: TSqlitePassSelectStmtTableDefs;
  { Prepared sql statements to speed up operations }
  InsertStmt: Pointer;
  UpdateStmt: Pointer;
  DeleteStmt: Pointer;
  InsertStmtText: String;
  UpdateStmtText: String;
  DeleteStmtText: String;
  FPrimaryKeyField: TSqlitePassSelectStmtFieldDef;
  { FieldDefs }
  FFieldDefs: TSqlitePassSelectStmtFieldDefs;
  public
  Constructor Create(Owner: TSqlitePassSelectStmtTableDefs);
  Destructor Destroy; override;
  procedure Assign(TableDef: TSqlitePassTableDef);
  procedure BindValues(const PreparedStmt: Pointer; const FieldsValues: Array of String; const BindPrimaryKey: Boolean);
  Property PrimaryKeyField: TSqlitePassSelectStmtFieldDef Read FPrimaryKeyField;
  Procedure PrepareSQLStmts;
  Procedure UnprepareSQLStmts;
  function PostDelete: Boolean;
  function PostUpdate: Boolean;
  function PostInsert: Boolean;
  property FieldDefs: TSqlitePassSelectStmtFieldDefs Read FFieldDefs;
  end;

TSqlitePassTableDef = class(TSqlitePassGenericTableDef)
  private
  { Owner }
  FTableDefs: TSqlitePassTableDefs;
  { InternalPrimaryKey }
  FPrimaryKeyField: TSqlitePassTableFieldDef;
  { FieldDefs }
  FFieldDefs: TSqlitePassTableFieldDefs;
  Public
  Constructor Create(Owner: TSqlitePassTableDefs);
  Destructor Destroy; override;
  procedure DeleteTable(Silent: Boolean = False);
  procedure RenameAs(NewName: String);
  procedure Reindex;
  property FieldDefs: TSqlitePassTableFieldDefs Read FFieldDefs;
  Property PrimaryKeyField: TSqlitePassTableFieldDef Read FPrimaryKeyField;
  end;

TSqlitePassTableDefs = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  Protected
  procedure ClearAndFreeItems;
  function GetAttached: Boolean;
  function GetItem(Index: Integer): TSqlitePassTableDef;
  procedure SetItem(Index: Integer; const Value: TSqlitePassTableDef);
  public
  constructor Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
  destructor Destroy; override;
  procedure CreateTable(Sql: String);
  procedure DeleteTable(TableName: String; Silent: Boolean = False);
  function TableByName(const Value: string): TSqlitePassTableDef;
  function FindTable(const Value: string): TSqlitePassTableDef;
  function FindFieldDef(TableName, FieldName: String): TSqlitePassTableFieldDef;
  function FindFieldTableName(FieldName: String; TableNames: TStringList): String;
  property Attached: Boolean Read GetAttached;
  property Count;
  procedure Refresh;
  property Items[Index: Integer]: TSqlitePassTableDef read GetItem write SetItem; default;
  end;


  { QueryDefs }

TSqlitePassQueryDef = class
  private
  FQueryName: String;
  FQueryFullName: String;
  FSql: String;
  FFieldDefs: TFieldDefs;
  FQueryDefs: TSqlitePassQueryDefs;
  FParams: Tparams;
  function GetAttached: Boolean;
  public
  Constructor Create(Owner: TSqlitePassQueryDefs);
  Destructor Destroy; override;
  property Attached: Boolean Read GetAttached;
  procedure DeleteQuery;
  procedure RenameAs(NewName: String);
  property QueryName: String Read FQueryName;
  property QueryFullName: String Read FQueryFullName;
  property Sql: String Read FSql;
  property Params: TParams Read FParams;
  property Fields: TFieldDefs Read FFieldDefs;
  end;

TSqlitePassQueryDefs = class(TSqlitePassObjectList)
  private
 

 FDatabase: TSqlitePassDatabase;
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  protected
  procedure ClearAndFreeItems;
  function GetAttached: Boolean;
  function GetItem(Index: Integer): TSqlitePassQueryDef;
  procedure SetItem(Index: Integer; const Value: TSqlitePassQueryDef);
  public
  constructor Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
  destructor Destroy; override;
  procedure CreateQuery(Sql: String);
  property Attached: Boolean Read GetAttached;
  procedure DeleteQuery(QueryName: String);
  function QueryByName(const Value: string): TSqlitePassQueryDef;
  function FindQuery(const Value: string): TSqlitePassQueryDef;
  property Count;
  procedure Refresh;
  property Items[Index: Integer]: TSqlitePassQueryDef read GetItem write SetItem; default;
  end;


{ Indexes management }

TSqlitePassIndexDirection = (cidAscending, cidDescending, cidUnknown);

TSqlitePassIndexColumn = class
 public
 Position: Integer;
 CollatingSequence: String;
 ColumnName: String;
 Direction: TSqlitePassIndexDirection;
 end;

TSqlitePassIndexColumns = class(TSqlitePassObjectList)
  private
  FIndex: TSqlitePassIndex;
  Protected
  procedure ClearAndFreeItems;
  function GetItem(Index: Integer): TSqlitePassIndexColumn;
  procedure SetItem(Index: Integer; const Value: TSqlitePassIndexColumn);
  public
  constructor Create(Owner: TSqlitePassIndex);
  destructor Destroy; override;
  function ColumnByName(const Value: string): TSqlitePassIndexColumn;
  function FindColumn(const Value: string): TSqlitePassIndexColumn;
  property Items[Index: Integer]: TSqlitePassIndexColumn read GetItem write SetItem; default;
  property Count;
  end;

TSqlitePassIndex = class
 private
  FDatasetName: String;
  FIndexName: String;
  FOverwriteIfExists: Boolean;
  FColumns : TSqlitePassIndexColumns;
  FSql: String;
  FUnique : Boolean;
  procedure SetFIndexName(value: String);
 public
  constructor Create;
  Destructor Destroy; override;
  property Sql: String Read FSql;
  property TableName: String Read FDatasetName Write FDatasetName;
  property IndexName: String Read FIndexName Write SetFIndexName;
  property Columns: TSqlitePassIndexColumns read FColumns write FColumns;
  property Unique: Boolean read FUnique write FUnique;
 end;

TSqlitePassDatabaseIndex = class(TSqlitePassIndex)
 private
  FIndexDefs: TSqlitePassDatabaseIndexDefs;
  function GetAttached: Boolean;
  public
  constructor Create(Owner: TSqlitePassDatabaseIndexDefs);
  Destructor Destroy; override;
  procedure Reindex;
  procedure DeleteIndex;
  property Attached: Boolean Read GetAttached;
  end;


TSqlitePassDatabaseIndexDefs = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  Protected
  procedure ClearAndFreeItems;
  function GetAttached: Boolean;
  function GetItem(Index: Integer): TSqlitePassIndex;
  procedure SetItem(Index: Integer; const Value: TSqlitePassIndex);
  public
  constructor Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
  destructor Destroy; override;
  function IndexByName(const Value: string): TSqlitePassIndex;
  function FindIndex(const Value: string): TSqlitePassIndex;
  procedure CreateIndex(Sql: String); overload;
  procedure CreateIndex(Index: TSqlitePassDatasetIndex); overload;
  procedure DeleteIndex(IndexName: String);
  procedure Reindex;
  procedure Refresh;
  procedure RenameIndex(OldIndexName, NewIndexName: String);
  property Attached: Boolean Read GetAttached;
  property Items[Index: Integer]: TSqlitePassIndex read GetItem write SetItem; default;
  property Count;
  end;


{ Views }

TSqlitePassView = class
 private
 FViewName: String;
 FTemp: Boolean;
 FSqlCreateStmt: String;
 FSqlSelectStmt: String;
 FViews: TSqlitePassViews;
 function GetAttached: Boolean;
 public
 Constructor Create(Owner: TSqlitePassViews);
 Destructor Destroy; override;
 procedure DeleteView;
 property Attached: Boolean Read GetAttached;
 property ViewName: String Read FViewName;
 property SqlCreateStmt: String Read FSqlCreateStmt;
 property SqlSelectStmt: String Read FSqlSelectStmt;
 end;


{ TSqlitePassViews }

TSqlitePassViews = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  Protected
  procedure ClearAndFreeItems;
  function GetAttached: Boolean;
  function GetItem(Index: Integer): TSqlitePassView;
  procedure SetItem(Index: Integer; const Value: TSqlitePassView);
  public
  constructor Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
  destructor Destroy; override;
  procedure CreateView(Sql: String);
  procedure DeleteView(ViewName: String);
  function ViewByName(const Value: string): TSqlitePassView;
  function FindView(const Value: string): TSqlitePassView;
  property Count;
  procedure Refresh;
  property Attached: Boolean Read GetAttached;
  property Items[Index: Integer]: TSqlitePassView read GetItem write SetItem; default;
  end;


{ Triggers }

TSqlitePassTrigger = class
 private
  FTemporary: Boolean;
  FTriggerName: String;
  FTriggers: TSqlitePassTriggers;
  FSql: String;
  function GetAttached: Boolean;
 public
  constructor Create(Owner: TSqlitePassTriggers);
  Destructor Destroy; override;
  procedure DeleteTrigger;
  property Attached: Boolean Read GetAttached;
  property Sql: String Read FSql;
  property Temporary: Boolean Read FTemporary Write FTemporary;
  property TriggerName: String Read FTriggerName Write FTriggerName;
 end;

TSqlitePassTriggers = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  Protected
  procedure ClearAndFreeItems;
  function GetAttached: Boolean;
  Function GetItem(Index: Integer): TSqlitePassTrigger;
  Procedure SetItem(Index: Integer; Value: TSqlitePassTrigger);
  public
  constructor Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
  destructor Destroy; override;
  procedure CreateTrigger(Sql: String);
  procedure DeleteTrigger(TriggerName: String);
  function TriggerByName(const Value: string): TSqlitePassTrigger;
  function FindTrigger(const Value: string): TSqlitePassTrigger;
  property Attached: Boolean Read GetAttached;
  property Items[Index: Integer]: TSqlitePassTrigger read GetItem write SetItem; default;
  property Count;
  procedure Refresh;
  end;


{TSqlitePassDatasets : Maintain a list of all datasets connected to the database}

TSqlitePassDatasets = Class(TSqlitePassObjectList)
  Private
    FDatabase: TSqlitePassDatabase;
  Protected
    Function GetItem(Index: Integer): TSqlitePassDataset;
    Procedure SetItem(Index: Integer; Dataset: TSqlitePassDataset);
  public
    constructor Create(Owner: TSqlitePassDatabase);
    destructor Destroy; override;
    function DatasetByName(Const Name: String): TSqlitePassDataset;
    function FindDataset(const Value: string): TSqlitePassDataset;
    property Items[Index: Integer]: TSqlitePassDataset read GetItem write SetItem; default;
    property Count;
    Procedure Close;
  end;

  TSqlitePassDatasetType = (dtTable, dtQuery, dtView, dtSQLDirect, dtUnknown);
 { Forward declaration }
 TSqlitePassDatasetIndexDefs=class;

 TSqlitePassDatasetVersionInfo = Class(TPersistent)
     Private
       FPackage     : String;
       FComponent   : String;
       procedure SetFComponent(const AValue: String);
       procedure SetFPackage(const AValue: String);
     Published
      constructor Create;
      Property Component     : String  Read FComponent       Write SetFComponent;
      Property Package       : String  Read FPackage         Write SetFPackage;
     end;


{ TSqlitePassMasterDataLink }

{ Used to synchronize fields.fieldNo with MemRecordset data }
TSqlitePassMasterDataLink = Class(TMasterDataLink)
 Protected
 procedure LayoutChanged; override;
 end;

{ TSqlitePassDataset }


TSqlitePassDatasetExportMethod        = (VisibleFields, AllFields);

{ Events }
TSqlitePassDatasetExportProgressEvent = procedure(Current, Total:Integer) of object;
TSqlitePassDatasetFilterRecordEvent = procedure(DataSet: TDataSet; var Accept: Boolean) of object;
TSqlitePassDatasetImportProgressEvent = procedure(Current:Integer) of object;
TSqlitePassDatasetGetFieldDataEvent = procedure(Field: TField; Const NativeFieldData: String; FieldDataBuffer: Pointer; var Handled: Boolean) of object;
TSqlitePassDatasetSetFieldDataEvent = Procedure(Field: TField; Const FieldDataBuffer: Pointer; Out NativeFieldData: String; var Handled: Boolean) of object;

TSqlitePassDataset = class(TDataSet)
private
    FRecordset: TSqlitePassRecordset;
    FParamCheck: Boolean;
    FParams: TParams;
    FDatasetName            : String;
    FDatasetFullName        : String;
    FDatasetType            : TSqlitePassDatasetType;
    FDatabase               : TSqlitePassDatabase;
    FFieldDefsInfo          : String; //Dummy string just used to enable display in object inspector
    { Locate }
    { Recordset used to find or locate specific records
     Used by (Locate, Lookup, Find...) }
    FLocateRecordset: TSqlitePassRecordset;
    FLocateSQL: TSqlitePassSelectStmt;
    FLocateTable: TList;
    { Filtering }
    FMasterLink             : TSqlitePassMasterDataLink;
    FMasterFields           : String;
    FDetailFields           : TList;
    FMasterAutoActivate     : Boolean;
    FFilterRecordLowerLimit : Cardinal;
    FFilterRecordUpperLimit : Cardinal;
    { }
    FReadOnly               : Boolean;
    FProcessMessages        : Boolean;
    FSQL                    : TStringList; { The original SQL statement }
    FSQLSelectStmt          : TSqlitePassSelectStmt; { Parsed SQL Statement to add filtering, sorting... capabilities }
    FSQLChanged             : Boolean; { Flag to indicate that FSQLSelectStmt needs to be updated }
    { Sorting }
    FSorted

                 : Boolean;
    FSortedBy               : String;
    {}
    FInInternalOpen         : Boolean; { Flag to indicate TSqlitePassDataset executes InternalOpen }
    FDatabaseAutoActivate   : Boolean;
    FVersionInfo            : TSqlitePassDatasetVersionInfo;
    FIndexDefs              : TSqlitePassDatasetIndexDefs;
    FExportMethod           : TSqlitePassDatasetExportMethod;

    { Additional Events }
    FOnGetFieldData         : TSqlitePassDatasetGetFieldDataEvent;
    FOnSetFieldData         : TSqlitePassDatasetSetFieldDataEvent;
    FOnExportProgress       : TSqlitePassDatasetExportProgressEvent;
    FOnImportProgress       : TSqlitePassDatasetImportProgressEvent;
    FOnFilterRecord         : TSqlitePassDatasetFilterRecordEvent;

    { --- }
    Function  GetActiveRecord:  TSqlitePassMemRecord;
    Procedure UpdateFSQL(Sender: TObject);
    Procedure UpdateFSQLStmt;
    Procedure SetFDatasetType;
    procedure SetFSQL(const Value: TStringList);
    Function  GetFDatabase:TSqlitePassDatabase;
    Procedure SetFDatabase(Value:TSqlitePassDatabase);
    Function  GetFDatasetName:String;
    Procedure SetFDatasetName(const Value:String);
    Procedure SetFReadOnly(const Value:Boolean);
    Function  GetFReadOnly:Boolean;
    Procedure SetFFilterRecordLowerLimit(const Value: Cardinal);
    Procedure SetFFilterRecordUpperLimit(const Value: Cardinal);
    procedure SetFSortedBy(const Value: String);
    procedure SetFSorted(const Value: Boolean);
    function  GetUniqueFieldName(FieldName: String): String;
    procedure SetFFieldDefsInfo(const Value: String);
    { ----- Filtering ----- }
    function  GetMasterDataSource: TDataSource;
    procedure SetMasterDataSource(Value: TDataSource);
    procedure MasterChanged(Sender: TObject);
    procedure MasterDisabled(Sender: TObject);
    function  GetMasterFields: string;
    procedure SetMasterFields(const Value: string);
    procedure ProcessMasterDetailFields;
protected
    { ----- Params ----- }
    Procedure SetParamsList(Value: TParams);
    Procedure UpdateParamsList;
    Procedure WriteParamData(Writer: TWriter);
    Function  GetParamsCount: Integer;
    Procedure DefineProperties(Filer: TFiler); override;
    Procedure ReadParamData(Reader: TReader);
    { ----- Override methods from TDataset ----- }
    Procedure Loaded; override;
    Procedure Notification(AComponent: TComponent; Operation: TOperation);Override;
    Function AllocRecordBuffer: PChar; override;
    Function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override; //Ok v1.00
    Function GetRecordSize: Word; override;
    Function IsCursorOpen: Boolean; override;
    Procedure ClearCalcFields(Buffer: PChar);override;
    Procedure FreeRecordBuffer(var Buffer: PChar); override;
    Procedure InternalOpen; override;
    Procedure InternalClose; override;
    Procedure InternalInitFieldDefs; override;
    Procedure InternalClearFieldDefs;
    Procedure InternalInitIndexDefs;
    { We need to Prefix TDataEvent because lazarus/fpc seems to have a bug :
      Without the 'db' prefix Lazarus gives the correct references when searching
      the inherited procedure but fpc gives error when compiling and seems to
      confuse TDataEvent with forms.TDataEvent }
    Procedure DataEvent(Event: db.TDataEvent; Info: Longint); override;
    Procedure InternalInitRecord(Buffer: PChar); override;
    Procedure InternalRefresh; override;

    Procedure SetFieldData(Field: TField; Buffer: Pointer);override;
    procedure SetFiltered(Value: Boolean); override;
    procedure SetFilterText(Const Value: String); override;
    Procedure UpdateIndexDefs; override;
    Function  GetCanModify: Boolean; override;
    { }
    Procedure CheckFiltered;
    function LocateRecord(const KeyFields: string; const KeyValues: Variant;
      Options: TLocateOptions; SyncCursor: Boolean): Boolean;
    Procedure RefreshActive;

    { ----- Navigation and Editing ----- }
    Procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    Procedure InternalCancel; override;
    Procedure InternalDelete; override;
    Procedure InternalEdit; override;
    Procedure InternalFirst;override;
    procedure InternalInsert; override;
    Procedure InternalLast;override;
    Procedure InternalPost; override;
    Procedure InternalSetToRecord(Buffer: PChar); override;
    { ----- Bookmarks ----- }
    Procedure InternalGotoBookmark(ABookmark: Pointer); override;
    // Procedure InternalMoveToBookmark(Bookmark: Pointer);
    Function  GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override; {From TDataset}
    Procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override; {From TDataset}
    // Keep the inherited - Function  GetBookmarkStr: TBookmarkStr; override;
    // Keep the inherited - Procedure SetBookmarkStr(const Value: TBookmarkStr); override;
    Procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    Procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    { ----- Filtering ----- }
    procedure DoOnNewRecord; override;
    Procedure UpdateFieldsNo;
    { ----- Miscalleous ---- }
    Procedure InternalHandleException; override;
    { ----- Records infos ----- }
    Function GetRecordCount  : Integer; override;
    Function GetRecNo        : Integer; override;
    Procedure SetRecNo        (Value: Integer); override;
    { ----- Blobs ---- }
    procedure AllocateBLOBPointers(Buffer: PChar);
    procedure FreeBlobPointers(Buffer: PChar);
    procedure FreeRecordPointers(Buffer: PChar);
  public
    { TMemo used as quick debug screen }
    {$IFDEF DEBUG}
    DebugScreen: TMemo;
    {$ENDIF}
    { ----- Constructor - Destructor ----- }
    Constructor Create(AOwner: TComponent); override;
    Destructor  Destroy; override;
    { ----- Bookmarks ----- }
    Function BookmarkValid(Bookmark: TBookmark): Boolean; override;
    { -----  ----- }
    Procedure Resync(Mode: TResyncMode);override;
    Function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    { ----- Blobs ---- }
    Function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
    procedure CloseBlob(Field: TField); override;
    { -----  ---- }
    Property ParamCount: Integer read GetParamsCount;
    Property DatasetType: TSqlitePassDatasetType Read FDatasetType;
    Property SQLSelectStmt: TSqlitePassSelectStmt Read FSQLSelectStmt;
    Procedure ApplyFilter;
    Procedure Sort;
    function Locate(const KeyFields: string; const KeyValues: Variant;
      Options: TLocateOptions): Boolean; override;
  published
    Property Database: TSqlitePassDatabase Read GetFDatabase Write SetFDatabase;
    Property ExportMethod: TSqlitePassDatasetExportMethod Read FExportMethod Write FExportMethod;
    property MasterFields: string read GetMasterFields write SetMasterFields;
    property MasterSource: TDataSource read GetMasterDataSource write SetMasterDataSource;
    property MasterSourceAutoActivate: Boolean read FMasterAutoActivate write FMasterAutoActivate;
    Property DatasetName: String Read GetFDatasetName Write SetFDatasetName;
    Property FilterRecordLowerLimit: Cardinal Read FFilterRecordLowerLimit Write SetFFilterRecordLowerLimit;
    Property FilterRecordUpperLimit: Cardinal Read FFilterRecordUpperLimit Write setFFilterRecordUpperLimit;
    Property SQL: TStringList Read FSQL Write SetFSQL;
    Property Sorted: Boolean Read FSorted Write SetFSorted;
    Property SortedBy: String Read FSortedBy Write SetFSortedBy;
    Property Params: TParams read FParams Write SetParamsList Stored False;
    Property FieldDefsInfo: String Read FFieldDefsInfo Write FFieldDefsInfo;
    Property IndexDefs: TSqlitePassDatasetIndexDefs  Read FIndexDefs Write FIndexDefs;
    Property ReadOnly: Boolean Read GetFReadOnly Write SetFReadOnly default False;
    Property DatabaseAutoActivate: Boolean Read FDatabaseAutoActivate  Write FDatabaseAutoActivate;
    Property VersionInfo: TSqlitePassDatasetVersionInfo Read FVersionInfo Write FVersionInfo;
    { Additionnal Events }
    Property OnGetFieldData: TSqlitePassDatasetGetFieldDataEvent Read FOnGetFieldData Write FOnGetFieldData;
    Property OnSetFieldData: TSqlitePassDatasetGetFieldDataEvent Read FOnGetFieldData Write FOnGetFieldData;
    Property OnExportProgress: TSqlitePassDatasetExportProgressEvent Read FOnExportProgress Write FOnExportProgress;
    Property OnImportProgress: TSqlitePassDatasetImportProgressEvent Read FOnImportProgress Write FOnImportProgress;
    Property OnFilterRecord  : TSqlitePassDatasetFilterRecordEvent read FOnFilterRecord write FOnFilterRecord;

    Property ParamCheck: Boolean Read FParamCheck Write FParamCheck;
    Property ProcessMessages: Boolean Read FProcessMessages Write FProcessMessages;
    { --- }
    Property BeforeOpen;
    Property AfterOpen;
    Property BeforeClose;
    Property AfterClose;
    Property BeforeInsert;
    Property AfterInsert;
    Property BeforeEdit;
    Property AfterEdit;
    Property BeforePost;
    Property AfterPost;
    Property BeforeCancel;
    Property AfterCancel;
    Property BeforeDelete;
    Property AfterDelete;
    Property BeforeScroll;
    Property AfterScroll;
    Property OnCalcFields;
    Property OnDeleteError;
    Property OnEditError;
    Property OnNewRecord;
    Property OnPostError;
    { --- }
    Property AutoCalcFields;
    Property Active;
    Property Filter;
    Property Filtered;
    Property FilterOptions;
  End;

{ ----- Blob fields stream ----- }

TSqlitePassBlobStream = class(TMemoryStream)
  private
    FField      : TBlobField;
    FDataSet    : TSqlitePassDataset;
    FOpened     : Boolean;
    FModified   : Boolean;
    FRecBlobStream : TMemoryStream;
  public
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode);
    destructor Destroy; override;
    Function Read(var Buffer; Count: Longint): Longint; override;
    Function Write(const Buffer; Count: Longint): Longint; override;
  End;

 TSqlitePassDatasetIndexDefs = class(TSqlitePassObjectList)
  private
  FDataset: TSqlitePassDataset;
  Protected
  function GetItem(Index: Integer): TSqlitePassIndex;
  procedure SetItem(Index: Integer; const Value: TSqlitePassIndex);
  public
  constructor Create(Owner: TSqlitePassDataset);
  destructor Destroy; override;
  procedure RefreshFromCache;
  property Items[Index: Integer]: TSqlitePassIndex read GetItem write SetItem; default;
  property Count;
  end;

 TSqlitePassDatasetIndex = class(TSqlitePassIndex)
 private
  FIndexDefs: TSqlitePassDatasetIndexDefs;
  public
  constructor Create(Owner: TSqlitePassDatasetIndexDefs);
  destructor Destroy; override;
  procedure Reindex;
  procedure DeleteIndex;
  end;

 TSqlitePassTranslator = class
  Private
    { Objects Reference }
    FDatabase: TSqlitePassDatabase;
    FFieldTypesTranslator: TSqlitePassFieldTypesTranslationRules;

    { Date/Time <-> String conversion utility functions }
    function _DateToStr(const Date: TDateTime): string; virtual;
    function _TimeToStr(const Time: TDateTime): string; virtual;
    function _DateTimeToStr(const DateTime: TDateTime): string; virtual;
    function _StrToDate(const S: string): TDateTime; virtual;
    function _StrToTime(const S: string): TDateTime; virtual;
    function _StrToDateTime(const S: string): TDateTime; virtual;

    { Float <-> String conversion utility functions }
    function _FloatToStr(const Float: Double): string; virtual;
    function _StrToFloat(const S: string): Double; virtual;

    { Boolean <-> String conversion utility functions }
    function _BooleanToStr(const BoolValue: Boolean): string; virtual;
    function _StrToBoolean(const S: string): Boolean; virtual;
  Public
    { --- Constructor / Destructor --- }
    Constructor Create(Owner: TSqlitePassDatabase);
    Destructor Destroy;override;

    { GetFieldData (Move Data from TField Value to Dataset internal Buffer) }
    function StrToBooleanFieldBuffer(Const StrValue: String; out Value: WordBool): Boolean;
    function StrToIntFieldBuffer(Const StrValue: String; out Value: Integer): Boolean;
    function StrToLargeIntFieldBuffer(Const StrValue: String; out Value: Int64): Boolean;
    function StrToFloatFieldBuffer(Const StrValue: String; out Value: Double): Boolean;
    function StrToDateFieldBuffer(Const StrValue: String; out Value: TDateTimeRec): Boolean;
    function StrToTimeFieldBuffer(Const StrValue: String; out Value: TDateTimeRec): Boolean;
    function StrToDateTimeFieldBuffer(StrValue: String; out Value: TDateTimeRec): Boolean;
    function StrToFieldBuffer(Const StrValue: String; out Value: Pointer): Boolean;
    function StrToStrFieldBuffer(Const StrValue: String; out Value: String): Boolean;
    function StrToByteFieldBuffer(Const StrValue: String; out Value: String): Boolean;

    { SetFieldData (Move Data from Dataset internal Buffer to TField Value) }
    procedure BooleanFieldBufferToStr(Const Value: Boolean; out StrValue: String);
    procedure IntFieldBufferToStr(Const Value: Integer; out StrValue: String);
    procedure LargeIntFieldBufferToStr(Const Value: Int64; out StrValue: String);
    procedure FloatFieldBufferToStr(Const Value: Double; out StrValue: String);
    procedure DateFieldBufferToStr(Const Value: TDateTimeRec; out StrValue: String);
    procedure TimeFieldBufferToStr(Const Value: TDateTimeRec; out StrValue: String);
    procedure DateTimeFieldBufferToStr(Const Value: TDateTimeRec; out StrValue: String);
    procedure FieldBufferToStr(Const Value: PChar; out StrValue: String);
    procedure ByteFieldBufferToStr(Const Value: PChar; out StrValue: String);

    { Binds Value to Sql statement before Dataset.post }
    function BindValue(const PreparedStmt: pointer; Const DataType: TFieldType; const FieldValue: String; const Idx: Integer): Boolean;
    function GetFieldData(Const StrValue: String; Const DataType: TFieldType; FieldBuffer: Pointer): Boolean;
    procedure SetFieldData(Const FieldBuffer: Pointer; Const DataType: TFieldType; out StrValue: String);

    { Gets information on database parts }
    procedure GetDatabaseTableDefs(TableDefs: TSqlitePassTableDefs); virtual;
    Procedure GetDatabaseTableFieldDefs(TableDef: TSqlitePassTableDef); virtual;
    procedure GetDatabaseIndexDefs(IndexDefs: TSqlitePassDatabaseIndexDefs); virtual;
    procedure GetDatabaseQueryDefs(QueryDefs: TSqlitePassQueryDefs); virtual;
    procedure GetDatabaseViews(Views: TSqlitePassViews); virtual;
    procedure GetDatabaseTriggers(Triggers: TSqlitePassTriggers); virtual;

    { Utilities }
    function IsSystemTable(TableName: String): Boolean; virtual;

    { Properties }
    property FieldTypesTranslator: TSqlitePassFieldTypesTranslationRules Read FFieldTypesTranslator;
  end;


  { Specific translator for Kexi databases }

  TSqlitePassTranslator_Kexi = class(TSqlitePassTranslator)
  private

   { Date/Time <-> String conversion utility functions }
    function _DateToStr(const Date: TDateTime): string; override;
    function _TimeToStr(const Time: TDateTime): string; override;
    function _DateTimeToStr(const DateTime: TDateTime): string; override;
    function _StrToDate(const S: string): TDateTime; override;
    function _StrToTime(const S: string): TDateTime; override;
    function _StrToDateTime(const S: string): TDateTime; override;

    { Boolean <-> String conversion utility functions }
    function _BooleanToStr(const BoolValue: Boolean): string; override;

    function ConvertNativeFieldTypeToPascalFieldType(FieldType: Word): TFieldType;
  public
    function GetNativeFieldTypeName(FieldType: Word): String;

    { Gets information on database parts }
    Procedure GetDatabaseTableFieldDefs(TableDef: TSqlitePassTableDef); override;
    procedure GetDatabaseQueryDefs(QueryDefs: TSqlitePassQueryDefs); override;

    { Utilities }
    function IsSystemTable(TableName: String): Boolean; override;
  end;

(*TSqlitePassTranslator_SqliteAdmin = class(TSqlitePassTranslator)
  private
    function GetKexiFieldType(FieldType: Word): TFieldType;
  public
    procedure InitFieldsTypes; override;
    procedure InitFieldDefs(Dataset: TSqlitePassDataset); override;
    procedure GetDatabaseObjects(ObjectType: TSqlitePassDatabaseObjectTypes); override;
    function IsSystemTable(TableName: String): Boolean; override;
  end;*)

StmtType = (stCreate, stCreateTable, stSelect, stUpdate, stDelete, stUnknown);

TSqlitePassTokenInfo = class
 public
 Index: Integer;
 TextStart: Integer;
 TextEnd: Integer;
 TextLength: Integer;
 Text: String;
 TextBeforeToken: String;
 TextAfterToken: String;
 Found: Boolean;
 end;

TSqlitePassSqlStmt = class
 private
  FInternalSQL: TStringList;
  FIsValid: Boolean;
  FStmtHandle: Pointer;
  function GetStmtType: StmtType;
 protected
  FDatabase: TSqlitePassDatabase;
  FSQL: TStringList;
  { Token utilities }
  procedure FindToken(Var StmtPart: String; StmtSchema: array of string; Token: TSqlitePassTokenInfo;
                      StartIndex: Integer = 0; DeleteStmtPart: Boolean = True);
  procedure InsertAfterToken(SectionName, TokenName, InsertedStmt: String);
  procedure RemoveAfterToken(SectionName, TokenName, RemovedStmt: String);
  procedure RemoveToken(Var StmtPart: String; Token: String);
  { Statement utilities }
  procedure InternalPrepare;
  procedure InternalUnPrepare;
  procedure InternalRebuildSQL(StmtKeywords: Array of string);
  procedure InternalSplitSections(StmtKeywords: Array of string);
  procedure CleanSQL;
  procedure QuoteString(Var S: String);
  procedure UnquoteString(Var S: String);
 public
  constructor Create(Database: TSqlitePassDatabase);
  destructor Destroy; override;
  { abstract functions and procedures }
  procedure RebuildSQL; virtual; abstract;
  procedure SplitSections; virtual; abstract;
  { }
  function  GetNextIdentifier(var StmtPart: String) : String;
  function FindSection(SectionName: String): String;
  procedure Prepare(Stmt: String); virtual;
  procedure ReplaceSection(SectionName: String; NewStmt: String);
  procedure UnPrepareSql; virtual;
  { Properties }
  property IsValid: Boolean Read FIsValid;
  property StmtHandle: Pointer Read FStmtHandle;
  property SQL: TStringList Read FSql;
  property SQLType: StmtType Read GetStmtType;
 end;

 (* NOT USED -> FOR FUTUR IMPLEMENTATION OF
   multiple SQL statements container in TSQLitePassDatabase *)
{ container for one or more sql statements }
(*TSqlitePassSqlStmts = class(TSqlitePassObjectList)
  private
  function GetItem(Index: Integer): TSqlitePassSqlStmt;
  procedure SetItem(Index: Integer; const Value: TSqlitePassSqlStmt);
  public
  constructor Create;
  destructor Destroy; override;
  property Count;
  property Items[Index: Integer]: TSqlitePassSqlStmt read GetItem write SetItem; default;
  end;*)


 TSqlitePassSelectStmt = class(TSqlitePassSqlStmt)
 private
  FPrimaryKeyStmt: String;
  FTableDefs: TSqlitePassSelectStmtTableDefs;
  FDataset: TSqlitePassDataset;
 protected
  procedure AddPrimaryKeys;
  function FillTableDefs(Const TableName, FieldName: String; FieldNo: Integer): Boolean;
  procedure

 PrepareSQLStmts;
  procedure UnprepareSQLStmts;
 public
  constructor Create(Dataset: TSqlitePassDataset);
  destructor Destroy; override;
  function BuildSchema: Boolean;
  procedure Prepare(Stmt: String); override;
  procedure UnPrepare; 
  procedure RebuildSQL; override;
  function Post(PostType: TSqlitePassPostType): Boolean;
  procedure SplitSections; override;
  property TableDefs: TSqlitePassSelectStmtTableDefs Read FTableDefs;
 end;

{ Internal representation of a record returned by a table/query }
TSqlitePassMemRecord = Class
    public
    RecordIndex: Integer;
    BookMarkFlag: TBookmarkFlag;
    FieldsValues: Array of String;
    constructor Create(FieldsCount: Integer);
    destructor Destroy(FieldDefs: TFieldDefs = nil); reintroduce;
    Procedure Assign(Const Source: TSqlitePassMemRecord; FieldDefs: TFieldDefs = nil);
    Procedure AssignFieldsValue(const Source: TSqlitePassMemRecord; FieldDefs: TFieldDefs = nil);
    Procedure ClearAndFree(FieldDefs: TFieldDefs = nil);
  end;

{ Internal CalcField List to quickly locate a Calc Field }
PCalcField = ^TCalcField;
TCalcField = record
     FieldName: String;
     FieldNo: Integer;
     end;

{ Internal cache holding the records returned by a table/query
  It is designed to provide data to TSqlitePassDataset }
TSqlitePassRecordset = Class(TSqlitePassObjectList)
 Private
    FActive: Boolean;
    FCurrentRecord: TSqlitePassMemRecord;
    FCurrentRecordOldValues: TSqlitePassMemRecord;
    FCurrentRecordIndex: Integer;
    FDatabaseHandle: Pointer;
    FFieldDefs: TFieldDefs;
    FFieldsNo: Array of integer;
    FSql: String;
    { --- Records --- }
    procedure SetFCurrentRecordIndex(const Value: Integer);
    procedure SetFCurrentRecord(const Value: TSqlitePassMemRecord);
    Procedure GetRecords;
    procedure UpdateRecordIndex(StartIndex: integer = 0);
    { --- Calculated fields utilities --- }
    procedure ClearCalcFieldsList;
    procedure InitCalcFieldsList(Fields: TFields);
    Function GetCalcFieldNo(FieldName: String): Integer;
    { --- --- }
  Protected
    CalcFieldsList: TList;
    FieldsCount: Cardinal;
    PreviousRecordIndex: Integer;
    NewRecord: TSqlitePassMemRecord;
    DataFieldsCount: Cardinal;
    Function GetItems(Index: Integer): TSqlitePassMemRecord;
    Procedure SetItems(Index: Integer; ARecord: TSqlitePassMemRecord);
  public
    Dataset: TSqlitePassDataset;
    MoveState: TGetResult;
    { --- Constructor / Destructor --- }
    constructor Create(Dataset: TSqlitePassDataset);
    Destructor Destroy; override;
    { --- Opening / Closing --- }
    Procedure Open(DatabaseHandle: Pointer; Sql: String);
    Procedure Close;
    { --- Navigating --- }
    Procedure First;
    Procedure Prior;
    Procedure Next;
    Procedure Last;
    Procedure MoveBy(Offset: Integer);
    { --- Managing Records --- }
    Procedure AddRecord;
    Function  Add(ARecord: TSqlitePassMemRecord): Integer;
    Procedure Insert(Index: Integer; ARecord: TSqlitePassMemRecord);
    Function  CreateNewRecord(Fields: TFields): TSqlitePassMemRecord;
    procedure Cancel;
    Procedure Delete(Index: Integer);
    { --- Current record old values storage / restore when editing --- }
    Procedure SaveCurrentRecordOldValues;
    Procedure RestoreCurrentRecordOldValues;
    { --- Calculated fields utilities --- }
    procedure SetFieldsNo(Fields: TFields); // TODO Remove from public
    Procedure ClearCalcFields(ARecord: TSqlitePassMemRecord);
    Function GetFieldNo(Index: Integer): Integer;
    { --- Properties --- }
    Property Active: Boolean Read FActive;
    Property CurrentRecord: TSqlitePassMemRecord Read FCurrentRecord Write SetFCurrentRecord;
    { CurrentRecordIndex is zero based }
    Property CurrentRecordIndex: Integer Read FCurrentRecordIndex Write SetFCurrentRecordIndex;
    Property Items[Index: Integer]: TSqlitePassMemRecord read GetItems write SetItems; default;
  end;

implementation
{$i SqlitePassDatabaseParts.inc}
{$i SqlitePassDatabase.inc}
{$i SqlitePassDataset.inc}
{$i SqlitePassSqlStmts.inc}
{$i SqlitePassRecordset.inc}
end.
