{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Basic sqlite library wrapper for use with fpc compiler
    Author : Luc DAVID Email: luckylazarus@free.fr
    Dowload : http
    portions of code from Mike Cariotoglou Email: mike@deltasingular.gr
    Copyright (C) 2006 - 2007

    Major changes are indicated in the \Documentation\Changes.pdf file
    Last update 10.01.2008

  --------------------------------------------------------------------------- }

unit SqlitePassEngine;
{$i SqlitePassDbo.inc}

{$DEFINE DEBUG}
interface

Uses
{$IFDEF FPC}
variants, LCLIntf,
{$ENDIF}
{$IFDEF WIN32}
Windows,
{$ELSE}
Dynlibs,
{$ENDIF}
SysUtils, Classes, Db, Forms, Controls,
SqlitePassApi_v3, SqlitePassErrorLang;

Const
{$IFDEF WIN32}
  {$DEFINE extdecl:=stdcall}
  DefaultSQLiteLibrary = 'sqlitepass3.dll';
{$ELSE}
  {$DEFINE extdecl:=cdecl}
  DefaultSQLiteLibrary = 'libsqlitepass3.so';
{$ENDIF}

Type
 PPointer= ^Pointer;
 TSqlitePassExecCallBack=function(acolumns: Integer; aColumnValues, aColumnNames:Ppchar):integer of object;
 PSqlitePassExecCallBack=^TSqlitePassExecCallBack;
 TSqlitePassGetPragmaCallback=function(UserData: Pointer; ColumnCount: Integer; ColumnValues, ColumnNames: PPointer): Integer; cdecl;

TSqlitePassEngine = class;


 TSqlitePassTransaction = class
 Private
  FLevel: Integer;
  FEngine: TSqlitePassEngine;
  FInternalTransaction: Boolean;
 Public
  Constructor Create(Owner: TSqlitePassEngine);
  Destructor Destroy; override;
  Procedure CommitInternalTransaction;
  Procedure RollbackInternalTransaction;
  Procedure StartInternalTransaction;
  Procedure ResetTransactions(CommitTransaction: Boolean);
  Procedure Commit;
  Procedure CommitAll;
  Procedure Rollback;
  Procedure RollbackAll;
  Procedure Start;
  Property  Level: Integer Read FLevel Write FLevel;
 end;

 TSqlitePassEngine = Class
  Private
    FPreparedStmts: TStringList;
    FConnectionHandle: Pointer;
    FLibraryFile: String;
    FLibraryHandle: Integer;
    FLibraryLoaded: Boolean;
    FStartTime:Integer;
    FStopTime: Integer;
    FElapsedTime : Double;
    FQueryTimeOut    : Integer;
    FTransaction: TSqlitePassTransaction;
  Public
    Constructor Create;
    Destructor Destroy; override;
    Property QueryTimeout: Integer Read FQueryTimeout Write FQueryTimeout;
    Property Connection: Pointer read FConnectionHandle;
    { Chrono }
    Procedure StartChrono;
    Procedure StopChrono;
    Property ElapsedTime: Double read FElapsedTime;
    { ------------ Library ----------------- }
    Function LoadSqliteLibrary(LibraryFile: String): Boolean;
    { ------------- Database ---------------- }
    procedure CreateDatabase(FullName: String);
    procedure OpenDatabase(FullName: String; LibraryFile: String = DefaultSQLiteLibrary);
    procedure CloseDatabase;
    { ----------- Transactions ------------- }
    Property Transaction: TSqlitePassTransaction Read FTransaction Write FTransaction;
    { -------- SQL Statements ----------- }
    function PrepareStmt(var Stmt: pointer; const sql: String): Boolean;
    procedure UnprepareStmt(var Stmt: pointer);
    Procedure FinalizePendingStmts;
    { ------------- Sql Exec ---------------- }
    function _ExecSql(Sql: string; CallBackFunc: TSqlitePassExecCallBack=nil): Integer;
    function Exec(Sql: string): Integer;
    function CheckResult(i: Integer; ShowMessage: Boolean = True): Boolean;
   end;



Implementation

{ ----- Utilities ----- }
function _ExecCallback(cbParam: PSqlitePassExecCallBack; Columns: Integer;
                       ColumnValues, ColumnNames: Ppchar): integer; cdecl;
begin
try
// Result:=cbparam^(columns,columnValues,columnNames);
except
 result:=1;
end;
end;


{ TSqliteEngine }

constructor TSqlitePassEngine.Create;
begin
FLibraryLoaded := False;
FLibraryHandle := 0;
FConnectionHandle:= nil;
FQueryTimeOut := 0;
FTransaction := TSqlitePassTransaction.Create(Self);
{$IFDEF DEBUG}
FPreparedStmts := TStringList.Create;
{$ENDIF}
end;

destructor TSqlitePassEngine.Destroy;
begin
CloseDatabase;
{$IFDEF DEBUG}
FPreparedStmts.Free;
{$ENDIF}
FTransaction.Free;
inherited Destroy;
end;

function TSqlitePassEngine.CheckResult(i: Integer; ShowMessage: Boolean = True): Boolean;
begin
Result := (i in [SQLITE_OK, SQLITE_DONE]);
If (Not Result) and  ShowMessage
   then raise Exception.CreateFmt('%d:%s',[i,SqliteDbv3_Errmsg(FConnectionHandle)])
end;

function TSqlitePassEngine._ExecSql(Sql: string; CallBackFunc: TSqlitePassExecCallBack): Integer;
begin
StartChrono;
repeat
 (*if Assigned(CallBackFunc)
    then result:= SqliteDbv3_Exec(FConnectionHandle,Pchar(Sql),@_ExecCallBack,@@CallBackFunc,nil)
    else*)
Result:= SqliteDbv3_Exec(FConnectionHandle,Pchar(Sql),nil,nil,nil);
until (result=SQLITE_OK) or not (result in [SQLITE_BUSY,SQLITE_LOCKED])
     or ((FQueryTimeOut>=0) and (Integer(GetTickCount)>=FStartTime));
StopChrono;
CheckResult(Result);
end;

function TSqlitePassEngine.Exec(Sql: string): Integer;
var
PSql, PRemainingSQL: Pchar;
DataPtr: Pointer;
begin
 StartChrono;
 PSql := PChar(Sql);
 repeat
  DataPtr:=nil;
  if CheckResult(SqliteDbv3_prepare_v2(FConnectionHandle,PSql,-1, DataPtr, PRemainingSQL), False) then
     Repeat Until not CheckResult(SqliteDbv3_step(DataPtr), False);

  { If we have another sql stmt, we continu }
  if PRemainingSQL^<>#0 then
   begin
   SqliteDbv3_finalize(DataPtr);
   PSql:=PRemainingSQL;
   end;
 Until (PRemainingSQL^= #0);

SqliteDbv3_Finalize(DataPtr);
StopChrono;
end;

Procedure TSqlitePassEngine.CreateDatabase(FullName:String);
var
NewDatabase: Pointer;
begin
if Not FileExists(FullName)
   then CheckResult(SqliteDbv3_Open(Pchar(FullName),NewDatabase));
end;

procedure TSqlitePassEngine.OpenDatabase(FullName: String; LibraryFile: String = DefaultSQLiteLibrary);
begin
FLibraryLoaded := LoadSqliteLibrary(LibraryFile);
CheckResult(SqliteDbv3_Open(Pchar(FullName),FConnectionHandle));
end;

procedure TSqlitePassEngine.CloseDatabase;
begin
If Assigned(FConnectionHandle) then
   begin
   Transaction.CommitAll;
   {$IFDEF DEBUG}
   FinalizePendingStmts;
   {$ENDIF}
   if CheckResult(SqliteDbv3_Close(FConnectionHandle)) then
      begin
      FConnectionHandle := nil;
      Dec(EnginesCount);
      end;
   end;
if FLibraryLoaded and (EnginesCount = 0)
   then FreeLibrary(FLibraryHandle);
end;

Function TSqlitePassEngine.LoadSqliteLibrary(LibraryFile: String): Boolean;
begin
Result := False;
if LibraryFile = '' then LibraryFile := DefaultSqliteLibrary;
{ The library is already loaded and we try to load the same file again}
if FLibraryLoaded and (LibraryFile = FLibraryFile) then exit;

if FLibraryLoaded then FreeLibrary(FLibraryHandle);


{$IFDEF WIN32}
FLibraryHandle := LoadLibrary(PChar(LibraryFile));
{$ELSE}
FLibraryHandle := LoadLibrary(LibraryFile);
{$ENDIF}

Result := FLibraryHandle <> 0;
if Result then Try
               LoadFunctions(FLibraryHandle);
               Except
               Result := False;
               end
          else DatabaseErrorFmt(Msg3001, [LibraryFile]);
end;

function TSqlitePassEngine.PrepareStmt(var Stmt: pointer; const sql: String): Boolean;
var
i: Integer;
RemainingSql: PChar;
begin
 UnprepareStmt(Stmt);
 Result := CheckResult(SqliteDbv3_prepare_v2(FConnectionHandle,Pchar(Sql),-1,Stmt,RemainingSql));
 {$IFDEF DEBUG}
 If Result then
    begin
    i := FPreparedStmts.IndexOfName(IntToStr(Integer(Stmt)));
    if i > -1
       then FPreparedStmts.Strings[i] := IntToStr(Integer(Stmt)) + '=' + Sql
       else FPreparedStmts.Add(IntToStr(Integer(Stmt)) + '=' + Sql);
    end;
 {$ENDIF}
end;

procedure TSqlitePassEngine.UnprepareStmt(var Stmt: pointer);
var
i: Integer;
begin
if FConnectionHandle = nil then exit;
if Assigned(Stmt)
   then begin
        try
        CheckResult(SqliteDbv3_Finalize(Stmt));
        {$IFDEF DEBUG}
        i := FPreparedStmts.IndexOfName(IntToStr(Integer(Stmt)));
        if i > -1
           then FPreparedStmts.Delete(i);
        {$ENDIF}
        finally
        Stmt := nil;
        end;
        end;
end;

{ TODO : Workaround to cleanup unused 'garbage' statements
  These statements should not exist, but...I have got to check my coding... }
procedure TSqlitePassEngine.FinalizePendingStmts;
var
Stmt: Pointer;
begin
While FPreparedStmts.Count > 0 do
  begin
  Stmt := Pointer(StrToInt(FPreparedStmts.Names[0]));
  UnPrepareStmt(Stmt);
  end;
end;

(*procedure TSqlitePassEngine.bindParamByName(const fst pointer; const param: tparam);
var idx:integer;
begin
 idx:=SqliteDbv3_bind_parameter_index(fst,pchar(param.Name));
 if idx=0 then raise exception.createFmt('parameter not found : %s',[param.Name]);
 BindParam(param,idx-1);
end;*)

{ TSqlitPassTransaction }

constructor TSqlitePassTransaction.Create(Owner: TSqlitePassEngine);
begin
inherited Create;
FEngine := Owner;
FLevel := 0;
end;

destructor TSqlitePassTransaction.Destroy;
begin
FEngine := nil;
Inherited Destroy;
end;

procedure TSqlitePassTransaction.Start;
begin
FEngine._ExecSql('Begin;');
Inc(FLevel);
end;

procedure TSqlitePassTransaction.Commit;
begin
FEngine._ExecSql('Commit;');
Dec(FLevel);
end;

procedure TSqlitePassTransaction.CommitAll;
begin
ResetTransactions(True);
end;

procedure TSqlitePassTransaction.Rollback;
begin
FEngine._ExecSql('Rollback;');
Dec(FLevel);
end;

procedure TSqlitePassTransaction.RollbackAll;
begin
ResetTransactions(False);
end;

procedure TSqlitePassTransaction.ResetTransactions(CommitTransaction: Boolean);
begin
While FLevel > 0 do
    Case CommitTransaction of
         True  : Commit;
         False : Rollback;
         end;
end;

procedure TSqlitePassTransaction.StartInternalTransaction;
begin
if FLevel = 0
   then begin
        FInternalTransaction := True;
        Start;
        end;
end;

procedure TSqlitePassTransaction.CommitInternalTransaction;
begin
if FInternalTransaction = True
   then begin
        FInternalTransaction := False;
        Commit;
        end;
end;

procedure TSqlitePassTransaction.RollbackInternalTransaction;
begin
if FInternalTransaction = True
   then begin
        FInternalTransaction := False;
        Rollback;
        end;
end;

procedure TSqlitePassEngine.StartChrono;
begin
 FStartTime:=Integer(GetTickCount);
end;

procedure TSqlitePassEngine.StopChrono;
begin
 FStopTime:=Integer(GetTickCount);
 FElapsedTime := (FStopTime - FStartTime)/1000;
end;

end.

