{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Basic SQL Parser and Statements Management
    Author : Luc DAVID Email: luckylazarus@free.fr
    Copyright (C) 2006 - 2007

    Major changes are indicated in the \Documentation\Changes.pdf file
    Last update 08.12.2007

  --------------------------------------------------------------------------- }

{ TSqlitePassSqlStmt }

constructor TSqlitePassSqlStmt.Create(Database: TSqlitePassDatabase);
begin
inherited Create;
FDatabase := Database;
FSQL := TStringList.Create;
FInternalSQL := TStringList.Create;
FStmtHandle := nil;
end;

destructor TSqlitePassSqlStmt.Destroy;
begin
UnprepareSql;
FInternalSQL.Free;
FSQL.Free;
FDatabase := nil;
Inherited Destroy;
end;

procedure TSqlitePassSqlStmt.Prepare(Stmt: String);
begin
FSQL.SetText(Pchar(Stmt));
SplitSections;
InternalPrepare;
end;

procedure TSqlitePassSqlStmt.InternalPrepare;
begin
 FIsValid := FDatabase.FEngine.PrepareStmt(FStmtHandle, PChar(SQL.Text));
end;

{ TODO: Was renamed to UnPrepareSql instead of UnPrepare (Virtual) because
  fpc compiler seems to be confused between TSqlitePassSqlStmt.Unprepare
  and TSqlitePassSelectStmt.Unprepare (At runtime, the wrong procedure is called)}
procedure TSqlitePassSqlStmt.UnPrepareSql;
begin
 FSQL.SetText('');
 InternalUnPrepare;
end;

procedure TSqlitePassSqlStmt.InternalUnPrepare;
begin
 if Assigned(FDatabase) and Assigned(FDatabase.FEngine) and Assigned(FStmtHandle) then
    begin
    FDatabase.Engine.UnprepareStmt(FStmtHandle);
    FStmtHandle := nil;
    end;
end;

function TSqlitePassSqlStmt.FindSection(SectionName: String): String;
begin
Result := FInternalSQL.Values[SectionName];
end;

procedure TSqlitePassSqlStmt.ReplaceSection(SectionName: String; NewStmt: String);
var
i: integer;
begin
i := FInternalSQL.IndexOfName(SectionName);
if i > -1
   then FInternalSQL[i] := SectionName + '=' + Trim(NewStmt);
RebuildSQL;
end;

procedure TSqlitePassSqlStmt.InternalSplitSections(StmtKeywords: array of string);
var
i, StartIndex: integer;
Stmt, SectionName, SectionText: String;
Token: TSqlitePassTokenInfo;
begin
Try
CleanSQL;
Stmt := FSQL.Text;
Token := TSqlitePassTokenInfo.Create;
StartIndex := 0;
Token.Index := StartIndex;
FInternalSQL.Clear;
{ Fills the StringList with empty sections }
For i := 0 to High(StmtKeywords) do
  FInternalSQL.Add(StmtKeywords[i]+'=');

{ splits sections }
Repeat
FindToken(Stmt, StmtKeywords, Token, StartIndex);
if Token.Found then
   begin
   SectionName := Token.Text;
   SectionText := Token.Text;
   StartIndex := Token.index + 1;
   FindToken(Stmt, StmtKeywords, Token, StartIndex, False);
   if Token.Found
      then SectionText := SectionText + kwSpace + Token.TextBeforeToken
      else SectionText := SectionText + kwSpace + Token.TextAfterToken;
   ReplaceSection(SectionName, SectionText);
   end;
until Not Token.Found;

finally
Token.Free;
RebuildSQL;
end;
end;

Procedure TSqlitePassSqlStmt.CleanSQL;
var
LastChar,i: integer;
OldSQL, NewSQL: String;
begin
OldSQL := Trim(FSQL.Text);
OldSQL := StringReplace(OldSQL, kwLineEnd, ' ', [rfIgnoreCase]);
LastChar := Length(OldSQL);
If LastChar = 0 then exit;
i := 1;
NewSQL := '';
Repeat
  NewSQL := NewSQL + OldSQL[i];
  Inc(I);
  { Trim extra spaces }
  While (OldSQL[i-1] = kwSpace) and (OldSQL[i] = kwSpace) do Inc(i)
Until (i > LastChar) or (OldSQL[i] = kwSqlEnd);
Dec(i);
{ Add ';' at the end if needed }
if OldSql[i] <> kwSqlEnd
   then NewSQL := NewSQL + kwSqlEnd;
FSQL.Text := NewSQL;
end;

Procedure TSqlitePassSqlStmt.InternalRebuildSQL(StmtKeywords: array of string);
var
i: integer;
SectionText: String;
begin
FSQL.Clear;
for i := 0 to High(StmtKeywords)-1 do
    begin
    SectionText := FindSection(StmtKeywords[i]);
    if SectionText <> '' then
       begin
       SectionText := StringReplace(SectionText, ';', '', []);
       FSQL.Add(SectionText);
       end;
    end;
{ Add ';' at the end if needed }
FSQL[Pred(FSQL.Count)] := FSQL[Pred(FSQL.Count)] + kwSqlEnd;
end;

procedure TSqlitePassSqlStmt.FindToken(var StmtPart: String;
  StmtSchema: array of string; Token: TSqlitePassTokenInfo; StartIndex: Integer = 0; DeleteStmtPart: Boolean = True);
var
i, TokenStart, TokenEnd, TokenLength: integer;
Found: Boolean;
PreviousChar, NextChar, UppercaseStmt: String;
begin
UppercaseStmt := Uppercase(StmtPart);
{ Clear Token info }
With Token do
  begin
  Index := -1;
  TextStart := 0;
  Text := '';
  TextLength := 0;
  TextEnd := 0;
  TextBeforeToken := '';
  TextAfterToken := StmtPart;
  { By default, search result is NotFound }
  Found := False;
  end;

for i := StartIndex to High(StmtSchema) do
  begin
  TokenStart := Pos(StmtSchema[i], UppercaseStmt);
  Found := (TokenStart > 0);

  { If one of the searched Token is found, we check if the token is not part of a larger identifier }
  if Found then
     begin
     if TokenStart = 1
        then PreviousChar := kwSpace
        else PreviousChar := UppercaseStmt[TokenStart-1];
     TokenLength := Length(StmtSchema[i]);
     TokenEnd := TokenStart + TokenLength-1;
     if TokenEnd = Length(UppercaseStmt)
        then NextChar := kwSpace
        else NextChar := UppercaseStmt[TokenEnd+1];

     Found := (PreviousChar[1] in BeforeIdentifierSeparator)
              and (NextChar[1] in AfterIdentifierSeparator);

     { If one of the requested token is found, we fill the data }
     if Found then With Token do
        begin
        Index := i;
        TextStart := TokenStart;
        Text := StmtSchema[i];
        TextLength := TokenLength;
        TextEnd := TokenEnd;
        TextBeforeToken := System.Copy(StmtPart, 1, TextStart-1);
        TextAfterToken := System.Copy(StmtPart, TextEnd+1, MaxInt);
        if DeleteStmtPart
           then System.Delete(StmtPart, 1, TextEnd);
        Found := True;
        Break;
        end;
     end;
  end;
end;

procedure TSqlitePassSqlStmt.InsertAfterToken(SectionName, TokenName,
  InsertedStmt: String);
var
Token: TSqlitePassTokenInfo;
SectionStmt: String;
begin
Token := TSqlitePassTokenInfo.Create;
SectionStmt := FindSection(SectionName);
FindToken(SectionStmt, TokenName, Token, 0, False);
if Token.Found then
   begin
   System.Insert(InsertedStmt, SectionStmt, Token.TextEnd+1);
   ReplaceSection(SectionName, SectionStmt);
   end;
Token.Free;
end;

procedure TSqlitePassSqlStmt.RemoveAfterToken(SectionName, TokenName,
  RemovedStmt: String);
var
Token: TSqlitePassTokenInfo;
SectionStmt: String;
begin
Token := TSqlitePassTokenInfo.Create;
SectionStmt := FindSection(SectionName);
FindToken(SectionStmt, TokenName, Token, 0, False);
if Token.Found then
   begin
   System.Delete(SectionStmt, Token.TextEnd+1, Length(RemovedStmt));
   ReplaceSection(SectionName, SectionStmt);
   end;
Token.Free;
end;

procedure TSqlitePassSqlStmt.RemoveToken(var StmtPart: String;
  Token: String);
var
StartPos: Integer;
UppercaseStmt : String;
begin
UppercaseStmt := Uppercase(StmtPart);
StartPos := Pos(Token, UppercaseStmt);
if StartPos > 0
   then System.Delete(StmtPart, StartPos, Length(Token));
end;

function TSqlitePassSqlStmt.GetNextIdentifier(var StmtPart: String) : String;
var
i, LastPos: Integer;
begin
Result := '';

{ Removes leading spaces }
i:=1;
While StmtPart[i] = kwSpace
    do System.Delete(StmtPart,1,1);

{ Extracts identifier until we find a separator char }
LastPos := Length(StmtPart);
For i := 1 to LastPos do
    begin
    if StmtPart[i] in AfterIdentifierSeparator
       then break;
    Result := Result + StmtPart[i];
    end;
Delete(StmtPart, 1, i-1);
UnquoteString(Result);
end;

procedure TSqlitePassSqlStmt.QuoteString(var S: String);
begin
If S <> '' then
   begin
   S := StringReplace(S, kwDot, FDatabase.Options.FQuoteOpen + kwDot + FDatabase.Options.FQuoteClose, [rfReplaceAll]);
   S := FDatabase.Options.FQuoteOpen + S + FDatabase.Options.FQuoteClose;
   end;
end;


Procedure TSqlitePassSqlStmt.UnquoteString(Var S: String);
begin
 if S <> '' then
    begin
    S := StringReplace(S, FDatabase.Options.FQuoteOpen, '', [rfReplaceAll]);
    S := StringReplace(S, FDatabase.Options.FQuoteClose, '', [rfReplaceAll]);
    end;
end;

function TSqlitePassSqlStmt.GetStmtType: StmtType;
var
Token: TSqlitePassTokenInfo;
Stmt: String;
begin
Token := TSqlitePassTokenInfo.Create;
Stmt := FSQL.Text;
FindToken(Stmt, StmtTypeKeywords, Token);
Case Token.Index of
     0   : Result := stCreate;
     1..3: Result := stCreateTable;
     4   : Result := stSelect;
     5   : Result := stUpdate;
     6   : Result := stDelete;
     else  Result := stUnknown;
     end;
end;



{ TSqlitePassSqlStmts }

(*constructor TSqlitePassSqlStmts.Create;
begin
inherited

end;

destructor TSqlitePassSqlStmts.Destroy;
begin
inherited
end;

function TSqlitePassSqlStmts.GetItem(Index: Integer): TSqlitePassSqlStmt;
begin
 Result:=TSqlitePassSqlStmt(Inherited Items[Index]);
end;

procedure TSqlitePassSqlStmts.SetItem(Index: Integer;
  const Value: TSqlitePassSqlStmt);
begin
 Put(Index,Value);
end;    *)


{ TSqlitePassSelectStmt }

constructor TSqlitePassSelectStmt.Create(Dataset: TSqlitePassDataset);
begin
if Assigned(Dataset.FDatabase)
   then Inherited Create(Dataset.FDatabase)
   else Inherited Create(nil);
FDataset := Dataset;
FTableDefs := TSqlitePassSelectStmtTableDefs.Create(Self);
FPrimaryKeyStmt := '';
end;

destructor TSqlitePassSelectStmt.Destroy;
begin
Unprepare;
FTableDefs.Free;
FDataset := nil;
inherited Destroy;
end;

{ Adds primary keys to a select statement if needed and then compiles the statement.
  Finally, calls PrepareSQLStmts to prepare and compile corresponding
  Insert-Update-Delete statements }
procedure TSqlitePassSelectStmt.Prepare(Stmt: String);
var
SelectStmt: String;
begin
FSQL.SetText(Pchar(Stmt));
if Stmt = '' then exit;
SplitSections;

{ Removes any previously added internal "rowid" primary keys from the select statemenent }
if FPrimaryKeyStmt <> '' then
   begin
   SelectStmt :=  FindSection(kwSelect);
   SelectStmt := StringReplace(SelectStmt, FPrimaryKeyStmt, '', [rfIgnoreCase]);
   FPrimaryKeyStmt := '';
   ReplaceSection(kwSelect, SelectStmt);
   end;

{ Gets the tables involved in the query in order to add Primary Keys to the statement
  BuildSchema is used to retrieve fieldDefs and information on PrimaryKey }
AddPrimaryKeys;

{ Rebuilds the definitive Schema including Primary Keys }
BuildSchema;

{ Prepares the SQL statements needed for inserting, updating or deleting fields data }
PrepareSQLStmts;

end;

Procedure TSqlitePassSelectStmt.UnPrepare;
begin
UnprepareSQLStmts;
Inherited UnPrepareSql;
end;

procedure TSqlitePassSelectStmt.SplitSections;
begin
InternalSplitSections(SelectStmtKeywords);
end;

procedure TSqlitePassSelectStmt.RebuildSQL;
begin
InternalRebuildSQL(SelectStmtKeywords);
end;


Procedure TSqlitePassSelectStmt.AddPrimaryKeys;
var
i, ColumnCount: Integer;
TableName: String;
TableNames: TStringList;
begin
Try
  InternalPrepare;
  if Not Assigned(StmtHandle) then exit;

  TableNames := TStringList.Create;
  TableNames.Sorted := True;
  TableNames.Duplicates := dupIgnore;

  ColumnCount := SqliteDbv3_column_count(StmtHandle);
  for i := 1 to ColumnCount do
       begin
       TableName := SqliteDbv3_column_table_name(StmtHandle, i-1);
       TableNames.Add(TableName);
       end;

  if TableNames.Count > 1
     then for i := 0 to Pred(TableNames.Count)
          do FPrimaryKeyStmt := FPrimaryKeyStmt + kwSpace + TableNames[i] + '.Rowid,'
     else FPrimaryKeyStmt := ' Rowid,';

  InsertAfterToken(kwSelect, kwSelect, FPrimaryKeyStmt);
  TableNames.Free;
finally
  InternalUnPrepare;
end;
end;

{ Builds a schema representation of the sql statement :
  A schema is made of tables and fieldDefs assigned from
  tables and fieldDefs info stored in the global database schema
}
function TSqlitePassSelectStmt.BuildSchema: Boolean;
var
i, ColumnCount: Integer;
TableName, FieldName: String;
begin
Result := False;

{ Clears the previous prepared schema }
UnprepareSQLStmts;
InternalUnPrepare;

{ Compiles the select statement }
InternalPrepare;
if Not Assigned(StmtHandle) then exit;

ColumnCount := SqliteDbv3_column_count(StmtHandle);
for i := 1 to ColumnCount do
     begin
     FieldName := SqliteDbv3_column_origin_name(StmtHandle, i-1);
     TableName := SqliteDbv3_column_table_name(StmtHandle, i-1);
     Result := FillTableDefs(TableName, FieldName, i);
     if Not Result then Exit
     end;

{ Checks for Primary Keys fields (Rowid) to mark them as the Internal Primary Key
  A statement is built like 'SELECT Table1.Rowid, Table2.Rowid, fieldnames... or * FROM...'
  so, the first FieldDef (FFieldDefs[0]) for a given TableDef will be his internal PrimaryKey }
For i := 0 to Pred(FTableDefs.Count) do
   With FTableDefs[i] do
   begin
   FFieldDefs[0].FInternalPrimaryKey := True;
   FFieldDefs[0].FFieldName := 'Rowid';
   FPrimaryKeyField := FFieldDefs[0];
   FPrimaryKeyFieldNo := FFieldDefs[0].FieldNo;
   end;
end;


{ The default column-name which identifies rows in SQLite is 'rowid',
  but for tables which have a primary integer key, this doesn't work
  (even though the documentation says it ought to). Therefore it is necessary
  to identify, for the given database, which tables need to use a different
  column-name for this. This gives a hash of table->column names. }
function TSqlitePassSelectStmt.FillTableDefs(Const TableName, FieldName: String; FieldNo: Integer): Boolean;
var
DatabaseTableDef: TSqlitePassTableDef;
StmtTableDef: TSqlitePassSelectStmtTableDef;
DatabaseFieldDef: TSqlitePassTableFieldDef;
StmtFieldDef: TSqlitePassSelectStmtFieldDef;

begin
Result := False;
  { Create a TableDef for a given table name }
   DatabaseTableDef := FDataset.Database.TableDefs.FindTable(TableName);
   if Assigned(DatabaseTableDef) then
      begin
      { if the TableDef doesn't already exist in select schema, we create it }
      StmtTableDef := TableDefs.FindTable(TableName);
      if StmtTableDef = nil then
         begin
         StmtTableDef := TSqlitePassSelectStmtTableDef.Create(FTableDefs);
         { Assigns doesn't fill FieldDefs }
         StmtTableDef.Assign(DatabaseTableDef);
         TableDefs.Add(StmtTableDef);
         end;
      DatabaseFieldDef := DatabaseTableDef.FieldDefs.FindFieldDef(FieldName);
      if Assigned(DatabaseFieldDef) then
         begin
        { a same FieldDef can be several times in a schema, so we always create one }
         StmtFieldDef := TSqlitePassSelectStmtFieldDef.Create(StmtTableDef.FFieldDefs);
         StmtFieldDef.Assign(DatabaseFieldDef);
         StmtFieldDef.FFieldNo := FieldNo;
         StmtTableDef.FieldDefs.Add(StmtFieldDef);
         Result := True;
         end;
      end;
end;

Procedure TSqlitePassSelectStmt.PrepareSQLStmts;
var
i: Integer;
begin
For i := 0 to Pred(FTableDefs.Count)
 do FTableDefs[i].PrepareSQLStmts;
end;

Procedure TSqlitePassSelectStmt.UnprepareSQLStmts;
var
i: Integer;
begin
For i := 0 to Pred(FTableDefs.Count)
 do FTableDefs[i].UnprepareSQLStmts;
 
FTableDefs.ClearAndFreeItems;
end;

function TSqlitePassSelectStmt.Post(PostType: TSqlitePassPostType): Boolean;
var
i: Integer;
begin
Result := False;
FDataset.FDatabase.FEngine.Transaction.StartInternalTransaction;
For i := 0 to Pred(FTableDefs.Count)
 do begin
    Case PostType of
         ptInsert: Result := FTableDefs[i].PostInsert;
         ptUpdate: Result := FTableDefs[i].PostUpdate;
         ptDelete: Result := FTableDefs[i].PostDelete;
         end;
    if Not Result then Break;
    end;
If Result
   then FDataset.FDatabase.FEngine.Transaction.CommitInternalTransaction
   else FDataset.FDatabase.FEngine.Transaction.RollbackInternalTransaction;
end;


{ TSqlitePassSelectStmtTableDefs }

constructor TSqlitePassSelectStmtTableDefs.Create(Owner: TSqlitePassSelectStmt);
begin
inherited Create;
FSQLStmt := Owner;
end;

destructor TSqlitePassSelectStmtTableDefs.Destroy;
begin
ClearAndFreeItems;
FSQLStmt := nil;
inherited Destroy;
end;

procedure TSqlitePassSelectStmtTableDefs.ClearAndFreeItems;
var
i: Integer;
TableDef: TSqlitePassSelectStmtTableDef;
begin
{ Clear the internal items }
for i := 0 to Pred(Count) do
  begin
  TableDef := GetItem(i);
  TableDef.Free;
  end;
{ Clear the list pointers }
inherited Clear;
end;


function TSqlitePassSelectStmtTableDefs.FindTable(
  const Value: string): TSqlitePassSelectStmtTableDef;
var
  I: Integer;
begin
  for I := 0 to Pred(Count) do
  begin
    Result := TSqlitePassSelectStmtTableDef(inherited Items[I]);
    if AnsiCompareText(Result.TableFullName, Value) = 0 then Exit;
  end;
  Result := nil;
end;

function TSqlitePassSelectStmtTableDefs.GetItem(Index: Integer): TSqlitePassSelectStmtTableDef;
begin
Result:=TSqlitePassSelectStmtTableDef(Inherited Items[Index]);
end;

procedure TSqlitePassSelectStmtTableDefs.SetItem(Index: Integer;
  const Value: TSqlitePassSelectStmtTableDef);
begin
  Put(Index,Value);
end;

{ TSqlitePassSelectStmtTableDef }

constructor TSqlitePassSelectStmtTableDef.Create(Owner: TSqlitePassSelectStmtTableDefs);
begin
 inherited Create;
 FTableDefs := Owner;
 FFieldDefs := TSqlitePassSelectStmtFieldDefs.Create(Self);
end;


destructor TSqlitePassSelectStmtTableDef.Destroy;
begin
 UnprepareSQLStmts;
 FPrimaryKeyField := nil;
 FFieldDefs.Free;
 FTableDefs := nil;
 inherited Destroy;
end;


procedure TSqlitePassSelectStmtTableDef.Assign(TableDef: TSqlitePassTableDef);
begin
 FAliasName := TableDef.AliasName;
 FCanModify := TableDef.CanModify;
 FOnConflict := TableDef.OnConlict;
 FPrimaryKeyField := nil;
 FPrimaryKeyFieldNo := -1;
 FSql := TableDef.Sql;
 FTableName := TableDef.TableName;
 FTableFullName := TableDef.TableFullName;
 FTemporary := TableDef.Temporary;
end;


{ Prepare standard Insert, Update, Delete sql statements to speed up
  engine operations. We will then just need to fill in the values using bindparams}
Procedure TSqlitePassSelectStmtTableDef.PrepareSQLStmts;
Var
i: Integer;
InsertStmtFields, InsertStmtValues, UpdateStmtFields: String;
begin
UnprepareSQLStmts;
{ Insert and Update statements}
InsertStmtFields := '';
InsertStmtValues := '';
UpdateStmtFields := '';
For i := 0 to Pred(FFieldDefs.Count)
    do if (not FFieldDefs[i].FinternalPrimaryKey) and (not FFieldDefs[i].FAutoInc) then
       begin
       InsertStmtFields := InsertStmtFields + FieldDefs[i].FieldName + ', ';
       InsertStmtValues := InsertStmtValues + ':'+ FieldDefs[i].FieldName + ', ';
       UpdateStmtFields := UpdateStmtFields + FieldDefs[i].FieldName + '= ?, ';
       end;
System.Delete(InsertStmtFields,Length(InsertStmtFields)-1,2);
System.Delete(InsertStmtValues,Length(InsertStmtValues)-1,2);
System.Delete(UpdateStmtFields,Length(UpdateStmtFields)-1,2);
InsertStmtText := 'INSERT INTO "' + TableFullName + '" (' + InsertStmtFields +') VALUES (' + InsertStmtValues + ');';
UpdateStmtText := 'UPDATE "' + TableFullName + '" SET ' + UpdateStmtFields + ' WHERE RowId=?;';
DeleteStmtText := 'DELETE FROM "' + TableFullName + '" WHERE RowId=?;';

With FTableDefs.FSQLStmt.FDatabase.FEngine do
     begin
     PrepareStmt(InsertStmt, InsertStmtText);
     PrepareStmt(UpdateStmt, UpdateStmtText);
     PrepareStmt(DeleteStmt, DeleteStmtText);
     end;
end;


Procedure TSqlitePassSelectStmtTableDef.UnprepareSQLStmts;
begin
{ Unprepare any previously prepared statement }
With FTableDefs.FSQLStmt.FDatabase.FEngine do
     begin
     UnprepareStmt(InsertStmt);
     UnprepareStmt(UpdateStmt);
     UnprepareStmt(DeleteStmt);
     end;
InsertStmtText := '';
UpdateStmtText := '';
DeleteStmtText := '';
end;


Function TSqlitePassSelectStmtTableDef.PostInsert: Boolean;
var
i: Integer;
InsertedRowId: String;
begin
 Result := False;
 BindValues(InsertStmt, FTableDefs.FSQLStmt.FDataset.FRecordset.NewRecord.FieldsValues, False);
 Result := FTableDefs.FSQLStmt.FDataset.FDatabase.FEngine.CheckResult(SqliteDbv3_step(InsertStmt), False);

 If Result then
    begin
    InsertedRowId := IntToStr(SqliteDbv3_last_insert_rowid(FTableDefs.FSQLStmt.FDataset.FDatabase.Engine.Connection));
    FTableDefs.FSQLStmt.FDataset.FRecordset.NewRecord.FieldsValues[FPrimaryKeyFieldNo-1] := InsertedRowId;
    // TODO : Speed up the primarykeys/primarykeys clones lookup
    For i := 0 to Pred(FFieldDefs.Count) do
        if FFieldDefs[i].FPrimaryKey or FFieldDefs[i].FAutoInc
           then FTableDefs.FSQLStmt.FDataset.FRecordset.NewRecord.FieldsValues[FFieldDefs[i].FieldNo-1] := InsertedRowId;
    //    FTableDefs.FSQLStmt.FDataset.FMemRecordset.FNewRecord.FieldsValues[FPrimaryKeyFieldNo] := InsertedRowId;
    end;
 SqliteDbv3_reset(InsertStmt);
end;

Function TSqlitePassSelectStmtTableDef.PostUpdate: Boolean;
begin
 Result := False;
 BindValues(UpdateStmt, FTableDefs.FSQLStmt.FDataset.FRecordset.CurrentRecord.FieldsValues, True);
 Result := FTableDefs.FSQLStmt.FDataset.FDatabase.FEngine.CheckResult(SqliteDbv3_step(UpdateStmt), False);
 SqliteDbv3_reset(UpdateStmt);
end;

Function TSqlitePassSelectStmtTableDef.PostDelete: Boolean;
begin
 Result := False;
 FTableDefs.FSQLStmt.FDataset.FDatabase.Translator.BindValue
 (DeleteStmt, ftInteger, FTableDefs.FSQLStmt.FDataset.FRecordset.CurrentRecord.FieldsValues[FPrimaryKeyFieldNo-1], 1);
 Result := FTableDefs.FSQLStmt.FDataset.FDatabase.FEngine.CheckResult(SqliteDbv3_step(DeleteStmt), False);
 SqliteDbv3_reset(DeleteStmt);
end;

procedure TSqlitePassSelectStmtTableDef.BindValues(const PreparedStmt: Pointer;
  const FieldsValues: Array of String; const BindPrimaryKey: Boolean);
var
i, ParamIndex: Integer;
begin
  FTableDefs.FSQLStmt.FDataset.FDatabase.FEngine.CheckResult(SqliteDbv3_clear_binding(PreparedStmt));
  ParamIndex := 1;

  for i:=0 to Pred(FFieldDefs.Count) do
   begin
   if(not FFieldDefs[i].FinternalPrimaryKey) and (not FFieldDefs[i].FAutoInc) then
      begin
      FTableDefs.FSQLStmt.FDataset.FDatabase.Translator.BindValue(PreparedStmt, FFieldDefs[i].DataType, FieldsValues[FFieldDefs[i].FieldNo-1], ParamIndex);
      Inc(ParamIndex);
      end;
   end;

  if BindPrimaryKey
     then FTableDefs.FSQLStmt.FDataset.FDatabase.Translator.BindValue(PreparedStmt, FPrimaryKeyField.DataType, FieldsValues[FPrimaryKeyFieldNo-1], ParamIndex);

end;


{ TSqlitePassSelectStmtFieldDef }

constructor TSqlitePassSelectStmtFieldDef.Create(
  Owner: TSqlitePassSelectStmtFieldDefs);
begin
 Inherited Create;
 FFieldDefs := Owner;
end;

destructor TSqlitePassSelectStmtFieldDef.Destroy;
begin
 FFieldDefs := nil;
 Inherited Destroy;
end;

procedure TSqlitePassSelectStmtFieldDef.Assign(FieldDef: TSqlitePassTableFieldDef);
begin
   FNoConstraints := FieldDef.NoConstraints;
   FAutoInc:= FieldDef.AutoInc;
   FDefaultValue := FieldDef.DefaultValue;
   FUnique := FieldDef.Unique;
   FPrimaryKey := FieldDef.PrimaryKey;
   FForeignKey := FieldDef.ForeignKey;
   FNotNull:= FieldDef.NotNull;
   FNotEmpty := FieldDef.NotEmpty;
   FIndexed := FieldDef.Indexed;
   FFieldName := FieldDef.FieldName;
   FFieldFullName := FieldDef.FieldFullName;
   FNativeDataType := FieldDef.NativeDataType;
   FNativeDataTypeCode := FieldDef.NativeDataTypeCode;
   FDataType := FieldDef.DataType;
   FDisplayName := FieldDef.DisplayName;
   FSize := FieldDef.Size;
   FPrecision:= FieldDef.Precision;
   FFieldNo := FieldDef.FieldNo;
   FHint := FieldDef.Hint;
end;

{ TSqlitePassSelectStmtFieldDefs }

constructor TSqlitePassSelectStmtFieldDefs.Create(
  Owner: TSqlitePassSelectStmtTableDef);
begin
 Inherited Create;
 FTableDef := Owner;
end;

destructor TSqlitePassSelectStmtFieldDefs.Destroy;
begin
 ClearAndFreeItems;
 FTableDef := nil;
 Inherited Destroy;
end;

procedure TSqlitePassSelectStmtFieldDefs.ClearAndFreeItems;
var
i: Integer;
FieldDef: TSqlitePassSelectStmtFieldDef;
begin
{ Clear the internal items }
for i := 0 to Pred(Count) do
  begin
  FieldDef := GetItem(i);
  FieldDef.Free;
  end;
{ Clear the list pointers }
inherited Clear;
end;

function TSqlitePassSelectStmtFieldDefs.GetItem(
  Index: Integer): TSqlitePassSelectStmtFieldDef;
begin
  Result:=TSqlitePassSelectStmtFieldDef(Inherited Items[Index]);
end;

procedure TSqlitePassSelectStmtFieldDefs.SetItem(Index: Integer;
  const Value: TSqlitePassSelectStmtFieldDef);
begin
  Put(Index,Value);
end;

procedure TSqlitePassSelectStmtFieldDefs.Assign(
  FieldDefs: TSqlitePassTableFieldDefs);
var
i: integer;
NewFieldDef: TSqlitePassSelectStmtFieldDef;
begin
for i := 0 to Pred(FieldDefs.Count) do
   begin
   NewFieldDef := TSqlitePassSelectStmtFieldDef.Create(Self);
   NewFieldDef.Assign(FieldDefs[i]);
   Self.Add(NewFieldDef);
   end;
end;





