{   This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   ---------------------------------------------------------------------------

    Messages localisation unit
    Author : Luc DAVID Email: luckylazarus@free.fr

   ---------------------------------------------------------------------------

    Utility unit : Uses mostly original code from FCL

      - Implement a TObjectList for Delphi 4
      - Implement TSqlitePassBitArray (Bit Array <-> Integer converter)


    Last update : 11.12.2007

   --------------------------------------------------------------------------- }
   
unit SqlitePassUtils;
{$i SqlitePassDbo.inc}

interface
Uses
{$IFDEF FPC}
 RtlConsts,
{$ELSE}
 Consts,
{$ENDIF}
{$IFDEF WIN32}
 Windows,
{$ENDIF}
Classes, SysUtils, Db;

Type
  TSqlitePassList = class(TObject)
  private
    FList: PPointerList;
    FCount: Integer;
    FCapacity: Integer;
  protected
    function Get(Index: Integer): Pointer;
    procedure Grow; virtual;
    procedure Put(Index: Integer; Item: Pointer);
    procedure SetCapacity(NewCapacity: Integer);
    function GetFCount: Integer;
    procedure SetCount(NewCount: Integer);
    //-------------------------------------------
    function Add(Item: Pointer): Integer;
    procedure Clear; dynamic;
    procedure Delete(Index: Integer);
    class procedure Error(const Msg: string; Data: Integer); virtual;
    procedure Exchange(Index1, Index2: Integer);
    function Expand: TSqlitePassList;
    function First: Pointer;
    function IndexOf(Item: Pointer): Integer;
    procedure Insert(Index: Integer; Item: Pointer);
    function Last: Pointer;
    procedure Move(CurIndex, NewIndex: Integer);
    procedure Pack;
    function Remove(Item: Pointer): Integer;
    procedure Sort(Compare: TListSortCompare);
    property Capacity: Integer read FCapacity write SetCapacity;
    property List: PPointerList read FList;
    public
    destructor Destroy; override;
    property Count: Integer read FCount write SetCount;
    property Items[Index: Integer]: Pointer read Get write Put; default;
    end;

  TSqlitePassObjectList = class(TSqlitePassList)
  private
    ffreeobjects : boolean;
  Protected
    function GetItem(Index: Integer): TObject;
    Procedure SetItem(Index: Integer; AObject: TObject);
    //--------------------------------------------------
    function IndexOf(AObject: TObject): Integer;
    Procedure Insert(Index: Integer; AObject: TObject);
    function First: TObject;
    Function Last: TObject;
    property Items[Index: Integer]: TObject read GetItem write SetItem; default;
  public
    // TODO
    function Add(AObject: TObject): Integer; // Moved to Public section to enable
                                             // access MyIndexDef.Columns.Add(MyIndexColumn;
                                             // demo program...
    function Remove(AObject: TObject): Integer; // Moved to Public section...same
  end;

  TSqlitePassBitArray = class
  private
  FIsZeroBased: Boolean;
  FItems: String;
  FInternalItems: TStringList;
  FMaxSize: Word;
  FOffset: Word;
  FValue: LongWord;
  FUpdating: Boolean;
  procedure CheckFirstItem;
  function  GetFValue: LongWord;
  procedure SetFValue(Value: LongWord);
  Procedure SetFIsZeroBased(Const Value: Boolean);
  Procedure SetFItems(Const Value: String);
  Procedure Reset;
  public
  constructor Create;
  destructor Destroy; override;
  function IsTrue(Const Item: String): Boolean;
  procedure SetItemValue(Const Item: String; Const Value: Boolean);
  property IsZeroBased: Boolean Read FIsZeroBased Write SetFIsZeroBased;
  property Items: String Read FItems Write SetFItems;
  property Value: LongWord Read GetFValue Write SetFValue;
  end;

  function GetExeDir: String;
  function GetWorkDir: String;
  function FieldTypeToString(FieldType: TFieldType): String;
  function StringToFieldType(FieldTypeName: String): TFieldType;
  function BoolAsString(Value: Boolean): String;
implementation


{ ----- Utilities ----- }

function GetExeDir: String;
begin
 SetLength(Result,1001);
 {$IFDEF WIN32}
 GetModuleFileName(HInstance,PChar(Result),1000);
 {$ENDIF};
  Result := ExtractFilePath(StrPas(PChar(Result)));
end;

function GetWorkDir: String;
begin
 GetDir(0, Result);
 if Result[Length(Result)] <> '\' then Result:=Result+'\';
end;

{ Show Datasets contents }

function FieldTypeToString(FieldType: TFieldType): String;
begin
Case FieldType of
     ftString      : Result := 'String';
     ftSmallint    : Result := 'Smallint';
     ftInteger     : Result := 'Integer';
     ftWord        : Result := 'Word';
     ftBoolean     : Result := 'Boolean';
     ftFloat       : Result := 'Float';
     ftCurrency    : Result := 'Currency';
     ftBCD         : Result := 'BCD';
     ftDate        : Result := 'Date';
     ftTime        : Result := 'Time';
     ftDateTime    : Result := 'DateTime';
     ftBytes       : Result := 'Bytes';
     ftVarBytes    : Result := 'VarBytes';
     ftAutoInc     : Result := 'AutoInc';
     ftBlob        : Result := 'Blob';
     ftMemo        : Result := 'Memo';
     ftGraphic     : Result := 'Graphic';
     ftFmtMemo     : Result := 'FmtMemo';
     ftParadoxOle  : Result := 'ParadoxOle';
     ftDBaseOle    : Result := 'DBaseOle';
     ftTypedBinary : Result := 'TypedBinary';
     ftCursor      : Result := 'Cursor';
     ftFixedChar   : Result := 'FixedChar';
     ftWideString  : Result := 'WideString';
     ftLargeint    : Result := 'Largeint';
     ftADT         : Result := 'ADT';
     ftArray       : Result := 'Array';
     ftReference   : Result := 'Reference';
     ftDataSet     : Result := 'DataSet';
     else
     Result := 'Unknown';
     end;
end;

function StringToFieldType(FieldTypeName: String): TFieldType;
var
i: TFieldType;
begin
Result := ftUnknown;
for i := Low(TFieldType) to High(TFieldType) do
    if FieldTypeName = FieldTypeNames[i] then
       begin
       Result := i;
       Break
       end;
end;
(*
if FieldTypeName = 'ftString'   then Result := ftString   else
if FieldTypeName = 'ftSmallint' then Result := ftSmallint else
if FieldTypeName = 'ftInteger'  then Result := ftInteger else
if FieldTypeName = 'ftWord'     then Result := ftWord else
if FieldTypeName = 'ftBoolean'  then Result := ftBoolean else
if FieldTypeName = 'ftFloat'    then Result := ftFloat else
if FieldTypeName = 'ftCurrency' then Result := ftCurrency else
if FieldTypeName = 'ftBCD'      then Result := ftBCD else
if FieldTypeName = 'ftDate'     then Result := ftDate else
if FieldTypeName = 'ftTime'     then Result := ftTime else
if FieldTypeName = 'ftDateTime' then Result := ftDateTime else
if FieldTypeName = 'ftBytes'    then Result := ftBytes else
if FieldTypeName = 'ftVarBytes' then Result := ftVarBytes else
if FieldTypeName = 'ftAutoInc'  then Result := ftAutoInc else
if FieldTypeName = 'ftBlob'     then Result := ftBlob else
if FieldTypeName = 'ftMemo'     then Result := ftMemo else
if FieldTypeName = 'ftGraphic'  then Result := ftGraphic else
if FieldTypeName = 'ftFmtMemo'  then Result := ftFmtMemo else
if FieldTypeName = 'ftParadoxOle' then Result := ftParadoxOle else
if FieldTypeName = 'ftDBaseOle' then Result := ftDBaseOle else
if FieldTypeName = 'ftTypedBinary' then Result := ftTypedBinary else
if FieldTypeName = 'ftCursor'    then Result := ftCursor else
if FieldTypeName = 'ftFixedChar' then Result := ftFixedChar else
if FieldTypeName = 'ftWideString' then Result := ftWideString else
if FieldTypeName = 'ftLargeint'  then Result := ftLargeint else
if FieldTypeName = 'ftADT'       then Result := ftADT else
if FieldTypeName = 'ftArray'     then Result := ftArray else
if FieldTypeName = 'ftReference' then Result := ftReference else
if FieldTypeName = 'ftDataSet'   then Result := ftDataSet
   else Result := ftUnknown; *)


function BoolAsString(Value: Boolean): String;
begin
Case Value of
     True  : Result := 'True';
     False : Result := 'False';
     end;
end;

{ TSqlitePassList }

destructor TSqlitePassList.Destroy;
begin
  Clear;
end;

function TSqlitePassList.Add(Item: Pointer): Integer;
begin
  Result := FCount;
  if Result = FCapacity then Grow;
  FList^[Result] := Item;
  Inc(FCount);
end;

procedure TSqlitePassList.Clear;
begin
  SetCount(0);
  SetCapacity(0);
end;

procedure TSqlitePassList.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  Dec(FCount);
  if Index < FCount then
    System.Move(FList^[Index + 1], FList^[Index],
      (FCount - Index) * SizeOf(Pointer));
end;

class procedure TSqlitePassList.Error(const Msg: string; Data: Integer);

  function ReturnAddr: Pointer;
  asm
          MOV     EAX,[EBP+4]
  end;

begin
  raise EListError.CreateFmt(Msg, [Data]) at ReturnAddr;
end;

procedure TSqlitePassList.Exchange(Index1, Index2: Integer);
var
  Item: Pointer;
begin
  if (Index1 < 0) or (Index1 >= FCount) then Error(SListIndexError, Index1);
  if (Index2 < 0) or (Index2 >= FCount) then Error(SListIndexError, Index2);
  Item := FList^[Index1];
  FList^[Index1] := FList^[Index2];
  FList^[Index2] := Item;
end;

function TSqlitePassList.Expand: TSqlitePassList;
begin
  if FCount = FCapacity then Grow;
  Result := Self;
end;

function TSqlitePassList.First: Pointer;
begin
  Result := Get(0);
end;

function TSqlitePassList.Get(Index: Integer): Pointer;
begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  Result := FList^[Index];
end;

procedure TSqlitePassList.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 64 then Delta := FCapacity div 4 else
    if FCapacity > 8 then Delta := 16 else
      Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

function TSqlitePassList.IndexOf(Item: Pointer): Integer;
begin
  Result := 0;
  while (Result < FCount) and (FList^[Result] <> Item) do Inc(Result);
  if Result = FCount then Result := -1;
end;

procedure TSqlitePassList.Insert(Index: Integer; Item: Pointer);
begin
  if (Index < 0) or (Index > FCount) then Error(SListIndexError, Index);
  if FCount = FCapacity then Grow;
  if Index < FCount then
    System.Move(FList^[Index], FList^[Index + 1],
      (FCount - Index) * SizeOf(Pointer));
  FList^[Index] := Item;
  Inc(FCount);
end;

function TSqlitePassList.Last: Pointer;
begin
  Result := Get(FCount - 1);
end;

procedure TSqlitePassList.Move(CurIndex, NewIndex: Integer);
var
  Item: Pointer;
begin
  if CurIndex <> NewIndex then
  begin
    if (NewIndex < 0) or (NewIndex >= FCount) then Error(SListIndexError, NewIndex);
    Item := Get(CurIndex);
    Delete(CurIndex);
    Insert(NewIndex, Item);
  end;
end;

procedure TSqlitePassList.Put(Index: Integer; Item: Pointer);
begin
  if (Index < 0) or (Index >= FCount) then Error(SListIndexError, Index);
  FList^[Index] := Item;
end;

function TSqlitePassList.Remove(Item: Pointer): Integer;
begin
  Result := IndexOf(Item);
  if Result <> -1 then Delete(Result);
end;

procedure TSqlitePassList.Pack;
var
  I: Integer;
begin
  for I := FCount - 1 downto 0 do if Items[I] = nil then Delete(I);
end;

procedure TSqlitePassList.SetCapacity(NewCapacity: Integer);
begin
  if (NewCapacity < FCount) or (NewCapacity > MaxListSize) then
    Error(SListCapacityError, NewCapacity);
  if NewCapacity <> FCapacity then
  begin
    ReallocMem(FList, NewCapacity * SizeOf(Pointer));
    FCapacity := NewCapacity;
  end;
end;

procedure TSqlitePassList.SetCount(NewCount: Integer);
begin
  if (NewCount < 0) or (NewCount > MaxListSize) then
    Error(SListCountError, NewCount);
  if NewCount > FCapacity then SetCapacity(NewCount);
  if NewCount > FCount then
    FillChar(FList^[FCount], (NewCount - FCount) * SizeOf(Pointer), 0);
  FCount := NewCount;
end;

procedure QuickSort(SortList: PPointerList; L, R: Integer;
  SCompare: TListSortCompare);
var
  I, J: Integer;
  P, T: Pointer;
begin
  repeat
    I := L;
    J := R;
    P := SortList^[(L + R) shr 1];
    repeat
      while SCompare(SortList^[I], P) < 0 do Inc(I);
      while SCompare(SortList^[J], P) > 0 do Dec(J);
      if I <= J then
      begin
        T := SortList^[I];
        SortList^[I] := SortList^[J];
        SortList^[J] := T;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(SortList, L, J, SCompare);
    L := I;
  until I >= R;
end;

procedure TSqlitePassList.Sort(Compare: TListSortCompare);
begin
  if (FList <> nil) and (Count > 0) then
    QuickSort(FList, 0, Count - 1, Compare);
end;

function TSqlitePassList.GetFCount: Integer;
begin
Result := FCount;
end;


{ TSqlitePassObjectList }

Function TSqlitePassObjectList.GetItem(Index: Integer): TObject;
begin
  Result:=TObject(Inherited Get(Index));
end;

Procedure TSqlitePassObjectList.SetItem(Index: Integer; AObject: TObject);
begin
  Put(Index,Pointer(AObject));
end;

Function TSqlitePassObjectList.Add(AObject: TObject): Integer;
begin
  Result:=Inherited Add(Pointer(AObject));
end;

Function TSqlitePassObjectList.Remove(AObject: TObject): Integer;
begin
  Result:=Inherited Remove(Pointer(AObject));
end;

Function TSqlitePassObjectList.IndexOf(AObject: TObject): Integer;
begin
  Result:=Inherited indexOF(Pointer(AObject));
end;

procedure TSqlitePassObjectList.Insert(Index: Integer; AObject: TObject);
begin
  Inherited Insert(Index,Pointer(AObject));
end;

function TSqlitePassObjectList.First: TObject;
begin
  Result := TObject(Inherited First);
end;

function TSqlitePassObjectList.Last: TObject;
begin
  Result := TObject(Inherited Last);
end;


{ TSqlitePassBitArray }

{ Utility object to manipulate an array of boolean values
  manipulated with a stringlist. This implementation is 'the easy way'
  and is surely not the fastest way to do it, but as we are limited up to
  32 (LongWord bit size) iterations, speed gain shouldn't be very important.

  Usage:
  1 - Create your TSqlitePassBitArray
    MyBooleanList := TSqlitePassBitArray.Create;

  2 - Set your list of items (number of items is limited to 32 in current
      implementation)
    MyBooleanList.Items := (Item1,Item2,Item3,Item4,Item5,Item6);
    or
    MyBooleanList.Items := (Item1;Item2;item3;item4;Item5;Item6);
      Item1=0
      Item2=0
      Item3=0
      Item4=0
      Item5=0
      Item6=0

  3 - Change the IsZeroBased propery if needed.
    if IsZeroBased is True
      then the first item of your list indicates whether
      or not ALL the others boolean values are false or if at
      least one item is set to True.
      In our example Item1=1 indicates that
      Item2=0
      Item3=0
      Item4=0
      Item5=0
      Item6=0

      In our example, we consider IsZeroBased = True,
      so Item1=0 indicates that at least one item is set to True :
      Item2=0
      Item3=0
      Item4=0
      Item5=1
      Item6=0

  4 - Set the value to set up the corresponding boolean state of each item
    MyBooleanList.Value := 11; ( 1011 in binary )
    will gives the resulting stringList :
      Item1=0 (at least one of the others items is set to True)
      --------
      Item2=1
      Item3=1
      Item4=0
      Item5=1
      --------
      Item6=0

  5 - Manipulate the values according to your needs :
    MyBooleanList.SetItemValue('Item6',True);
    MyBooleanList.SetItemValue('Item3',False);
    will give
      Item1=0 (at least one of the others items is set to True)
      Item2=1
      Item3=0
      Item4=0
      Item5=1
      Item6=1

  6 - Retrieve the integer representation of your boolean values
      MyValueToStore := MyBooleanList.Value;
      The value returned is 25 (1+8+16) }

constructor TSqlitePassBitArray.Create;
begin
FInternalItems:=TStringList.Create;
FUpdating := False;
FMaxSize := (SizeOf(Value)*8)-1;
IsZeroBased := True;
end;

destructor TSqlitePassBitArray.Destroy;
begin
FInternalItems.Free;
end;

{ We convert the Boolean values of the stringlist
  to a LongWord. This value will be easily stored in a
  field or in any integer variable }
function TSqlitePassBitArray.GetFValue: LongWord;
var
i: Integer;
begin
Result := 0;
for i := Pred(FInternalItems.Count) Downto FOffset do
    if FInternalItems.Values[FInternalItems.Names[i]] = '1'
       then Result := Result + (1 shl (i-Foffset));
end;

{ Returns True if the Item value is True otherwise returns False }
function TSqlitePassBitArray.IsTrue(Const Item: String): Boolean;
var
Pos: Integer;
begin
Try
Result := Boolean(StrToInt(FInternalItems.Values[Item]));
Except
Result := False;
End;
end;

{ Change the boolean value of an item }
procedure TSqlitePassBitArray.SetItemValue(Const Item: String; Const Value: Boolean);
var
i: Integer;
begin
i := FInternalItems.IndexOfName(Item);
if i > -1 then
   begin
   FInternalItems.Strings[i] := FInternalItems.Names[i] + '=' + IntToStr(Integer(Value));
   If not FUpdating then CheckFirstItem;
   end;
end;

{ Set the global value. This value is converted to binary
  and every item boolean value is updated }
procedure TSqlitePassBitArray.SetFValue(Value: LongWord);
var
i: Word;
begin
FValue := Value;
Reset;
i := FOffset;
While Value > 0 do
  begin
  SetItemValue(FInternalItems.Names[i], (Value mod 2) = 1);
  Value := Value div 2;
  Inc(i)
  end;
end;

{ Create the list and the items }
procedure TSqlitePassBitArray.SetFItems(Const Value: String);
begin
FInternalItems.BeginUpdate;
try
FItems := Value;
Reset;
If not FUpdating then CheckFirstItem;
finally
FInternalItems.EndUpdate;
end;
end;

procedure TSqlitePassBitArray.SetFIsZeroBased(const Value: Boolean);
begin
FIsZeroBased := Value;
if FIsZeroBased
   then FOffset := 1
   else FOffset := 0;
// Recalc the items boolean values
SetFValue(FValue);
end;

procedure TSqlitePassBitArray.CheckFirstItem;
var
i: integer;
begin
Try
{ Internal flag to avoid SetItemValue and CheckFirstItem
  to be called recursively }
FUpdating := True;

if IsZeroBased
   then begin
   SetItemValue(FInternalItems.Names[0], True);
   for i := FOffset to Pred(FInternalItems.Count) do
   if IsTrue(FinternalItems.Names[i])
      then begin
           SetItemValue(FInternalItems.Names[0], False);
           Exit;
           end;
   end;
finally
FUpdating := False;
end;
end;

procedure TSqlitePassBitArray.Reset;
var
Start, i, FItemsLength: integer;
S: string;
begin
FInternalItems.Clear;
i := 1;
FItemsLength := Length(FItems);
While i < FItemsLength do
    begin
    Start := i;
    While (not (FItems[i] in [',' , ';'])) and (i <= FItemsLength) do inc(i);
    s := System.Copy(FItems, Start, i-Start);
    FInternalItems.Add(System.Copy(FItems, Start, i-Start)+'=0');
    Inc(i);
    end;
end;

end.
